# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import os
import shutil
from threading import Lock
from zipfile import ZipFile, ZIP_DEFLATED, BadZipfile

from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.storage.h5.decorators import synchronized
from tvb.storage.h5.file.exceptions import FileStructureException
from tvb.storage.h5.file.xml_metadata_handlers import XMLWriter, XMLReader

LOCK_CREATE_FOLDER = Lock()


class FilesHelper(object):
    """
    This class manages all Structure related operations, using File storage.
    It will handle creating meaning-full entities and retrieving existent ones.
    """
    PROJECTS_FOLDER = "PROJECTS"
    ALLEN_MOUSE_CONNECTIVITY_CACHE_FOLDER = "ALLEN_MOUSE_CONNECTIVITY_CACHE"
    TUMOR_DATASET_FOLDER = "TUMOR_DATASET"

    def __init__(self):
        self.logger = get_logger(self.__class__.__module__)

    ############# PROJECT RELATED methods ##################################

    @synchronized(LOCK_CREATE_FOLDER)
    def check_created(self, path):
        """
        Check that the given folder exists, otherwise create it, with the entire tree of parent folders.
        This method is synchronized, for parallel access from events, to avoid conflicts.
        """
        try:
            # if this is meant to be used concurrently it might be better to catch OSError 17 then checking exists
            if not os.path.exists(path):
                self.logger.debug("Creating folder:" + str(path))
                os.makedirs(path, mode=TvbProfile.current.ACCESS_MODE_TVB_FILES)
                os.chmod(path, TvbProfile.current.ACCESS_MODE_TVB_FILES)
        except OSError:
            self.logger.exception("COULD NOT CREATE FOLDER! CHECK ACCESS ON IT!")
            raise FileStructureException("Could not create Folder" + str(path))

    @staticmethod
    def get_projects_folder():
        base_path = TvbProfile.current.TVB_STORAGE
        if TvbProfile.current.web.ENCRYPT_STORAGE and TvbProfile.current.web.CAN_ENCRYPT_STORAGE and TvbProfile.current.web.DECRYPT_PATH:
            base_path = TvbProfile.current.web.DECRYPT_PATH

        return os.path.join(base_path, FilesHelper.PROJECTS_FOLDER)

    def get_project_folder(self, project_name, *sub_folders):
        """
        Retrieve the root path for the given project.
        If root folder is not created yet, will create it.
        """
        complete_path = os.path.join(self.get_projects_folder(), project_name)
        if sub_folders is not None:
            complete_path = os.path.join(complete_path, *sub_folders)
        if not os.path.exists(complete_path):
            self.check_created(complete_path)
        return complete_path

    @staticmethod
    def get_project_folder_from_h5(h5_file):
        op_folder = os.path.dirname(h5_file)
        return os.path.dirname(op_folder)

    def rename_project_structure(self, project_name, new_name):
        """ Rename Project folder or THROW FileStructureException. """
        try:
            path = self.get_project_folder(project_name)
            folder = os.path.split(path)[0]
            new_full_name = os.path.join(folder, new_name)

            if os.path.exists(new_full_name):
                raise IOError("Path exists %s " % new_full_name)
            os.rename(path, new_full_name)
            return path, new_full_name
        except Exception:
            self.logger.exception("Could not rename node!")
            raise FileStructureException("Could not rename to %s" % new_name)

    def get_project_meta_file_path(self, project_name, project_file):
        """
        Retrieve project meta info file path.
        
        :returns: File path for storing Project meta-data
            File might not exist yet, but parent folder is created after this method call.
            
        """
        complete_path = self.get_project_folder(project_name)
        complete_path = os.path.join(complete_path, project_file)
        return complete_path

    @staticmethod
    def read_project_metadata(project_path, project_file):
        project_cfg_file = os.path.join(project_path, project_file)
        return XMLReader(project_cfg_file).read_metadata_from_xml()

    @staticmethod
    def write_project_metadata_from_dict(project_path, meta_entity, project_file):
        project_cfg_file = os.path.join(project_path, project_file)
        XMLWriter(meta_entity).write_metadata_in_xml(project_cfg_file)
        os.chmod(project_path, TvbProfile.current.ACCESS_MODE_TVB_FILES)

    def write_project_metadata(self, meta_dictionary, project_file):
        """
        :param meta_dictionary: Project metadata
        :param project_file: Project file path
        """
        proj_path = self.get_project_folder(meta_dictionary['name'])
        self.write_project_metadata_from_dict(proj_path, meta_dictionary, project_file)

    ############# OPERATION related METHODS Start Here #########################

    def remove_operation_data(self, project_name, operation_id):
        """
        Remove H5 storage fully.
        """
        try:
            complete_path = self.get_project_folder(project_name, str(operation_id))
            self.logger.debug("Removing: " + str(complete_path))
            self.remove_folder(complete_path, False)
        except FileStructureException:
            self.logger.exception("Could not remove files")
            raise FileStructureException("Could not remove files for OP" + str(operation_id))

    ####################### DATA-TYPES METHODS Start Here #####################
    def move_datatype(self, new_project_name, new_op_id, full_path):
        """
        Move H5 storage into a new location
        """
        try:
            folder = self.get_project_folder(new_project_name, str(new_op_id))
            full_new_file = os.path.join(folder, os.path.split(full_path)[1])
            os.rename(full_path, full_new_file)
        except Exception:
            self.logger.exception("Could not move file")
            raise FileStructureException("Could not move " + full_path)

    ######################## IMAGES METHODS Start Here #######################    
    def get_images_folder(self, project_name, images_folder):
        """
        Computes the name/path of the folder where to store images.
        """
        project_folder = self.get_project_folder(project_name)
        images_folder = os.path.join(project_folder, images_folder)
        self.check_created(images_folder)
        return images_folder

    def write_image_metadata(self, figure, meta_entity, images_folder):
        """
        Writes figure meta-data into XML file
        """
        _, dict_data = figure.to_dict()
        XMLWriter(meta_entity).write_metadata_in_xml(self._compute_image_metadata_file(figure, images_folder))

    def remove_figure(self, figure, images_folder):
        """
        Remove the file storing image and its meta data
        """
        figures_folder = self.get_images_folder(figure.project.name, images_folder)
        path2figure = os.path.join(figures_folder, figure.file_path)
        metadata_file = self._compute_image_metadata_file(figure, images_folder)
        self.remove_files([path2figure, metadata_file])

    def _compute_image_metadata_file(self, figure, images_folder):
        """
        Computes full path of image meta data XML file. 
        """
        name = figure.file_path.split('.')[0]
        images_folder = self.get_images_folder(figure.project.name, images_folder)
        return os.path.join(TvbProfile.current.TVB_STORAGE, images_folder, name + XMLWriter.FILE_EXTENSION)

    def get_allen_mouse_cache_folder(self, project_name):
        project_folder = self.get_project_folder(project_name)
        folder = os.path.join(project_folder, self.ALLEN_MOUSE_CONNECTIVITY_CACHE_FOLDER)
        self.check_created(folder)
        return folder

    def get_tumor_dataset_folder(self):
        project_folder = self.get_projects_folder()
        folder = os.path.join(os.path.dirname(project_folder), self.TUMOR_DATASET_FOLDER)
        self.check_created(folder)
        return folder

        ######################## GENERIC METHODS Start Here #######################

    @staticmethod
    def copy_file(source, dest, dest_postfix, buffer_size):
        """
        Copy a file from source to dest. source and dest can either be strings or 
        any object with a read or write method, like StringIO for example.
        """
        should_close_source = False
        should_close_dest = False

        try:
            if not hasattr(source, 'read'):
                source = open(source, 'rb')
                should_close_source = True

            if not hasattr(dest, 'write'):
                if dest_postfix is not None:
                    dest = os.path.join(dest, dest_postfix)
                if not os.path.exists(os.path.dirname(dest)):
                    os.makedirs(os.path.dirname(dest))
                dest = open(dest, 'wb')
                should_close_dest = True

            shutil.copyfileobj(source, dest, length=buffer_size)

        finally:
            if should_close_source:
                source.close()
            if should_close_dest:
                dest.close()

    @staticmethod
    def remove_files(file_list, ignore_exception=False):
        """
        :param file_list: list of file paths to be removed.
        :param ignore_exception: When True and one of the specified files could not be removed, an exception is raised.  
        """
        for file_ in file_list:
            try:
                if os.path.exists(file_):
                    if os.path.isfile(file_):
                        os.remove(file_)
                    if os.path.isdir(file_):
                        shutil.rmtree(file_)
            except Exception:
                logger = get_logger(__name__)
                logger.exception("Could not remove " + str(file_))
                if not ignore_exception:
                    raise

    @staticmethod
    def remove_folder(folder_path, ignore_errors):
        """
        Given a folder path, try to remove that folder from disk.
        :param ignore_errors: When False throw FileStructureException if folder_path is invalid.
        """
        if os.path.isdir(folder_path):
            shutil.rmtree(folder_path, ignore_errors)
            return
        if not ignore_errors:
            raise FileStructureException("Given path does not exists, or is not a folder " + str(folder_path))

    @staticmethod
    def compute_size_on_disk(file_path):
        """
        Given a file's path, return size occupied on disk by that file.
        Size should be a number, representing size in KB.
        """
        if os.path.isfile(file_path):
            return int(os.path.getsize(file_path) / 1024)
        return 0

    @staticmethod
    def compute_recursive_h5_disk_usage(start_path):
        """
        Computes the disk usage of all h5 files under the given directory.
        :param start_path:
        :return: A tuple of size in kiB
        """
        total_size = 0
        n_files = 0
        for dir_path, _, file_names in os.walk(start_path):
            for f in file_names:
                if f.endswith('.h5'):
                    fp = os.path.join(dir_path, f)
                    total_size += os.path.getsize(fp)
                    n_files += 1
        return int(round(total_size / 1024.))


class TvbZip(ZipFile):
    def __init__(self, dest_path, mode):
        try:
            ZipFile.__init__(self, dest_path, mode, ZIP_DEFLATED, True)
        except Exception as excep:
            raise FileStructureException("Could not open the given ZIP file..." + str(excep))
        self.logger = get_logger(self.__class__.__module__)

    def __enter__(self):
        return self

    def __exit__(self, _type, _value, _traceback):
        self.close()

    def write_zip_folder(self, folder, exclude, need_parent_folder=False):
        """
        write folder contents in archive
        :param archive_path_prefix: root folder in archive. Defaults to "" the archive root
        :param exclude: a list of file or folder names that will be recursively excluded
        :param need_parent_folder: if it is True, the parent_folder will be added as well to the zip file name
        :
        """
        parent_folder = folder
        if need_parent_folder:
            parent_folder, _ = os.path.split(folder)

        for root, dirs, files in os.walk(folder):
            for ex in exclude:
                ex = str(ex)
                if ex in dirs:
                    dirs.remove(ex)
                if ex in files:
                    files.remove(ex)

            for file_n in files:
                abs_file_n = os.path.join(root, file_n)
                zip_file_n = abs_file_n[len(parent_folder) + len(os.sep):]
                self.write(abs_file_n, zip_file_n)

    def write_zip_folders(self, folders, exclude):
        """
        This method creates a ZIP file with all folders provided as parameters
        :param folders: array with the FULL names/path of the folders to add into ZIP
        :param folder_prefix: root folder in archive. Defaults to "" the archive root
        :param exclude: a list of file or folder names that will be recursively excluded
        """
        for folder in set(folders):
            self.write_zip_folder(folder, exclude, True)

    def unpack_zip(self, folder_path):
        """ Simple method to unpack ZIP archive in a given folder. """

        def to_be_excluded(name):
            excluded_paths = ["__MACOSX/", ".DS_Store"]
            for excluded in excluded_paths:
                if name.startswith(excluded) or name.find('/' + excluded) >= 0:
                    return True
            return False

        try:
            result = []
            for filename in self.namelist():
                if not to_be_excluded(filename):
                    result.append(self.extract(filename, folder_path))
            return result
        except BadZipfile as excep:
            self.logger.exception("Could not process zip file")
            raise FileStructureException("Invalid ZIP file..." + str(excep))
        except Exception as excep:
            self.logger.exception("Could not process zip file")
            raise FileStructureException("Could not unpack the given ZIP file..." + str(excep))



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
This module contains logic for meta-data handling.

It handles read/write operations in XML files for retrieving/storing meta-data.
More specific: it contains XML Reader/Writer Utility, for generic metadata dictionary.

.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>

"""

import json
import xml.dom.minidom
from xml.dom.minidom import Node, Document

from tvb.basic.profile import TvbProfile
from tvb.basic.logger.builder import get_logger


class XMLReader(object):
    """
    Reader for XML with meta-data on generic entities (e.g. Project, Operation).
    """

    def __init__(self, xml_path):
        self.logger = get_logger(self.__class__.__module__)
        self.xml_path = xml_path

    def read_metadata_from_xml(self):
        """
        Return a dictionary, filled with data read from XML file.
        """
        self.logger.debug("Starting to parse XML file " + self.xml_path)
        root_node = self._find_root()
        # Parse all nodes, and read text content.    
        result_data = self._parse_xml_node_to_dict(root_node)
        return result_data

    @staticmethod       
    def get_node_text(node):
        """
        From XMl node, read string content.
        """
        for text_child in node.childNodes:
            if text_child.nodeType == Node.TEXT_NODE:
                return str(text_child.data).lstrip().rstrip()
        
        return ''

    ####### PRIVATE METHODS Start Here #######################################

    def _find_root(self):
        """
        From given file path, get XML root node.
        """
        doc_xml = xml.dom.minidom.parse(self.xml_path)
        for child_node in doc_xml.childNodes:
            if child_node.nodeType == Node.ELEMENT_NODE:
                return child_node
        return None

    def _parse_xml_node_to_dict(self, root_node):
        """
        Parse a given input XML node, and return the dictionary of text attributes.
        The dictionary can have multiple levels of deepness, but when child-nodes are 
        encountered, text value of the node is ignored.
        """
        result = {}
        for node in root_node.childNodes:
            if node.nodeType == Node.ELEMENT_NODE:
                result[node.nodeName] = self.get_node_text(node)
                result_meta = self._parse_xml_node_to_dict(node)
                if list(result_meta) is not None and len(list(result_meta)) > 0:
                    result[node.nodeName] = result_meta
        return result
    

class XMLWriter(object):
    """
    Writer for XML with meta-data on generic entities (e.g. Project, Operation).
    """
    ELEM_ROOT = "tvb_data"
    FILE_EXTENSION = ".xml"

    def __init__(self, entity):
        """
        :param entity:  a dictionary that contains generic metadata to be written
        """
        self.entity = entity
        self.logger = get_logger(self.__class__.__module__) 

    def write_metadata_in_xml(self, final_path):
        """
        From a meta-data dictionary for an entity, create the XML file.
        """
        doc = Document()
        root_node = doc.createElement(self.ELEM_ROOT)
        
        # Add information about version of the stored data
        root_node.setAttribute(TvbProfile.current.version.DATA_VERSION_ATTRIBUTE,
                               str(TvbProfile.current.version.DATA_VERSION))
        
        # Add each generic metadata attribute as XML node-elements.
        for att_name, att_value in self.entity.items():
            node = doc.createElement(att_name)
            if isinstance(att_value, list):
                att_value = json.dumps(att_value)
            node.appendChild(doc.createTextNode(str(att_value)))
            root_node.appendChild(node)
        doc.appendChild(root_node)

        # Now dump the XML content into a file.
        with open(final_path, 'wt') as file_obj:
            doc.writexml(file_obj, addindent="\t", newl="\n")



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
    Module used to test storage of TVB data into HDF5 format.
    
    .. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
"""

import os
import numpy
import shutil
import pytest

from tvb.basic.profile import TvbProfile
from tvb.storage.h5.file.exceptions import MissingDataSetException, IncompatibleFileManagerException, \
    FileStructureException
from tvb.storage.h5.file.hdf5_storage_manager import HDF5StorageManager
from tvb.storage.storage_interface import StorageInterface

# Some constants used by tests

STORAGE_FILE_NAME = "test_data.h5"
DATASET_NAME_1 = "dataset1"
DATASET_NAME_2 = "dataset2"
META_KEY = "meta_key"
META_VALUE = "meta_value"
META_DICT = {META_KEY: META_VALUE}
STORE_PATH = "/node1/node2"


class TestHDF5Storage(object):
    """
    This tests storage of data into HDF5 format (H5 files).
    """

    def setup_method(self):
        """
        Set up the context needed by the tests.
        """
        self.storage_folder = os.path.join(TvbProfile.current.TVB_TEMP_FOLDER, "test_hdf5")

        if os.path.exists(self.storage_folder):
            shutil.rmtree(self.storage_folder)
        os.makedirs(self.storage_folder)

        # Now create HDF5 storage instance
        self.storage = HDF5StorageManager(os.path.join(self.storage_folder, STORAGE_FILE_NAME))

        self.test_2D_array = numpy.random.random((10, 10))
        self.test_3D_array = numpy.random.random((3, 3, 3))
        self.test_string_array = numpy.array(
            [["a".encode('utf-8'), "b".encode('utf-8')], ["c".encode('utf-8'), "d".encode('utf-8')]])

    def teardown_method(self):
        """
        Tear down to revert any changes made by a test.
        """
        self.storage.close_file()

        if os.path.exists(self.storage_folder):
            shutil.rmtree(self.storage_folder)

    @staticmethod
    def _assert_arrays_are_equal(expected_arr, current_arr, message=None):
        numpy.testing.assert_array_equal(expected_arr, current_arr, message)

    # -------------- TESTS ---------------------
    def test_invalid_storage_path(self):
        """
        This method will test scenarios where no storage path or storage file is provided
        """
        self.storage_folder = os.path.join(TvbProfile.current.TVB_TEMP_FOLDER, "test_hdf5")
        # Test if folder name is None
        with pytest.raises(FileStructureException):
            HDF5StorageManager(None)

    def test_file_creation(self):
        """
        Test if the storage file is created on disk.
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        full_path = os.path.join(self.storage_folder, STORAGE_FILE_NAME)
        assert os.path.exists(full_path), "Storage file not created."

    def test_simple_data_storage(self):
        """
        Test if simple array data is stored
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        # Now read data
        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)

        self._assert_arrays_are_equal(self.test_2D_array, read_data, "Did not get the expected data")

    def test_store_data_repeatedly(self):
        """
        Test if simple array data is stored
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        ## Update with the same shape should work:
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        try:
            ## But update with a different shape should throw an exception
            self.storage.store_data(self.test_3D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
            raise AssertionError("Update with a different shape is expected to fail")
        except IncompatibleFileManagerException:
            pass

        # Now read data
        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_2D_array, read_data, "Did not get the expected data")

    def test_simple_data_storage_on_path(self):
        """
        Test if simple array data is stored on a given path
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, STORE_PATH)
        # Now read data
        read_data = self.storage.get_data(DATASET_NAME_1, None, STORE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_2D_array, read_data)

    def test_string_data_storage(self):
        """
        Test store of string data (array of strings
        """
        self.storage.store_data(self.test_string_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        # Now read data
        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_string_array, read_data)

    def test_store_none_data(self):
        """
        Test scenario when trying to store None data
        """
        with pytest.raises(FileStructureException):
            self.storage.store_data(None, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

    def test_append_data1(self):
        """
        Test data store using append method
        """
        self.storage.append_data(self.test_3D_array, DATASET_NAME_1, 1, True, StorageInterface.ROOT_NODE_PATH)
        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_3D_array, read_data)

    def test_append_data2(self):
        """
        Test data store using append method (multiple calls)
        """
        for index in range(self.test_2D_array.shape[-1]):
            slices = (slice(None, None, 1), slice(index, index + 1, 1))

            self.storage.append_data(self.test_2D_array[slices], DATASET_NAME_1, 1, True,
                                     StorageInterface.ROOT_NODE_PATH)

        self.storage.close_file()
        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_2D_array, read_data)

    def test_append_data_on_path(self):
        """
        Test data store using append method on a given path
        """
        self.storage.append_data(self.test_3D_array, DATASET_NAME_1, 1, True, STORE_PATH)
        read_data = self.storage.get_data(DATASET_NAME_1, None, STORE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_3D_array, read_data)

    def test_append_other_dimension(self):
        """
        This method test append operation on other dimension
        """
        for index in range(self.test_2D_array.shape[0]):
            slices = (slice(index, index + 1, 1), slice(None, None, 1))

            self.storage.append_data(self.test_2D_array[slices], DATASET_NAME_1, 0, True,
                                     StorageInterface.ROOT_NODE_PATH)

        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_2D_array, read_data)

    def test_append_strings(self):
        """
        Test appending strings 
        """
        for index in range(self.test_string_array.shape[-1]):
            slices = (slice(None, None, 1), slice(index, index + 1, 1))

            self.storage.append_data(self.test_string_array[slices], DATASET_NAME_1, 1, True,
                                     StorageInterface.ROOT_NODE_PATH)

        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_string_array, read_data)

    def test_append_none_data(self):
        """
        Test appending null value to dataset
        """
        with pytest.raises(FileStructureException):
            self.storage.append_data(None, DATASET_NAME_1, 1, True, StorageInterface.ROOT_NODE_PATH)

    def test_append_without_closing_file(self):
        """
        Test appending data but keeping hdf5 file open. 
        """
        for index in range(self.test_3D_array.shape[-1]):
            slices = (slice(None, None, 1), slice(index, index + 1, 1), slice(None, None, 1))
            self.storage.append_data(self.test_3D_array[slices], DATASET_NAME_1, 1, False,
                                     StorageInterface.ROOT_NODE_PATH)

        self.storage.close_file()

        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_3D_array, read_data)

    def test_close_file_multiple_time(self):
        """
        Test closing H5 file multiple times.
        """
        for index in range(self.test_3D_array.shape[-1]):
            slices = (slice(None, None, 1), slice(index, index + 1, 1), slice(None, None, 1))
            self.storage.append_data(self.test_3D_array[slices], DATASET_NAME_1, 1, False,
                                     StorageInterface.ROOT_NODE_PATH)

        self.storage.close_file()
        self.storage.close_file()

        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_3D_array, read_data)

    def test_delete_dataset(self):
        """
        This test checks deletion of a dataset from H5 file.
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        # Now check if data was persisted. 
        read_data = self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_2D_array, read_data)

        # Now delete dataset
        self.storage.remove_data(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        with pytest.raises(MissingDataSetException):
            self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)

    def test_delete_dataset_on_path(self):
        """
        This test checks deletion of a dataset from H5 file, dataset which is stored 
        under a given path
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, STORE_PATH)
        # Now check if data was persisted. 
        read_data = self.storage.get_data(DATASET_NAME_1, None, STORE_PATH, False, True)
        self._assert_arrays_are_equal(self.test_2D_array, read_data)

        # Now delete dataset
        self.storage.remove_data(DATASET_NAME_1, STORE_PATH)
        with pytest.raises(MissingDataSetException):
            self.storage.get_data(DATASET_NAME_1, None, StorageInterface.ROOT_NODE_PATH, False, True)

    def test_delete_missing_dataset(self):
        """
        This test checks deletion of non existing  dataset.
        """
        with pytest.raises(FileStructureException):
            self.storage.remove_data(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

    def test_read_sliced_data(self):
        """
        This test checks reading of data on slices.
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        # Now check if data was persisted, but read it using slices
        for index in range(len(self.test_2D_array)):
            sl = slice(index, index + 1, 1)
            read_data = self.storage.get_data(DATASET_NAME_1, sl, StorageInterface.ROOT_NODE_PATH, False, True)
            self._assert_arrays_are_equal(self.test_2D_array[sl], read_data)

    def test_add_metadata(self):
        """
        This method checks metadata add for root or a dataset
        """
        # Create a dataset 
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        # Now add meta info to it.
        self.storage.set_metadata(META_DICT, DATASET_NAME_1, True, StorageInterface.ROOT_NODE_PATH)
        read_meta_value = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_value[META_KEY]

        # Now we'll test adding metadata on root node
        self.storage.set_metadata(META_DICT, '', True, StorageInterface.ROOT_NODE_PATH)
        read_meta_value = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_value[META_KEY]

    def test_add_metadata_on_path(self):
        """
        This method checks metadata add for a dataset stored under a given path
        """
        # Create a dataset 
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, STORE_PATH)

        # Now add meta info to it.
        self.storage.set_metadata(META_DICT, DATASET_NAME_1, True, STORE_PATH)
        read_meta_value = self.storage.get_metadata(DATASET_NAME_1, STORE_PATH)
        assert META_VALUE == read_meta_value[META_KEY]

    def test_delete_metadata(self):
        """
        Test deletion of metadata for a dataset or root node.
        """
        # Create a dataset 
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        # Now add meta info to it.
        self.storage.set_metadata(META_DICT, DATASET_NAME_1, True, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_data[META_KEY]
        self.storage.remove_metadata(META_KEY, DATASET_NAME_1, True, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert 0 == len(read_meta_data), "There should be no metadata stored on dataset"

        # Now we'll test removal of metadata on root node
        self.storage.set_metadata(META_DICT, '', True, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_data[META_KEY], "Retrieved meta value is not correct"
        self.storage.remove_metadata(META_KEY, '', True, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert 1 == len(read_meta_data), "There should be no metadata stored on root node, except data version"

    def test_delete_metadata_on_path(self):
        """
        Test deletion of metadata for a dataset stored on a given path
        """
        # Create a dataset 
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, STORE_PATH)

        # Now add meta info to it.
        self.storage.set_metadata(META_DICT, DATASET_NAME_1, True, STORE_PATH)
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, STORE_PATH)
        assert META_VALUE == read_meta_data[META_KEY]
        self.storage.remove_metadata(META_KEY, DATASET_NAME_1, True, STORE_PATH)
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, STORE_PATH)
        assert 0 == len(read_meta_data), "There should be no metadata stored on dataset"

    def test_delete_missing_metadata(self):
        """
        This method tests removal of a missing metadata
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        # Test delete on a node
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert 0 == len(read_meta_data), "There should be no metadata stored on dataset"

        # Test delete on root node
        read_meta_data = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert 1, len(read_meta_data) == "There should be no metadata stored on root node, except data version"

    def test_delete_metadata_missing_dataset(self):
        """
        Test retrieval of metadata for a missing dataset.
        """
        # Create
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        with pytest.raises(MissingDataSetException):
            self.storage.get_metadata(DATASET_NAME_2, StorageInterface.ROOT_NODE_PATH)

    def test_concurent_file_access(self):
        """
        This method tests scenario when HDF5 file is opened concurrent for read & write
        """
        new_storage = HDF5StorageManager(os.path.join(self.storage_folder, STORAGE_FILE_NAME))
        new_storage.store_data(self.test_2D_array, DATASET_NAME_2, StorageInterface.ROOT_NODE_PATH)

        for index in range(self.test_3D_array.shape[-1]):
            slices = (slice(None, None, 1), slice(index, index + 1, 1), slice(None, None, 1))
            # Write into file, but leave file open
            self.storage.append_data(self.test_3D_array[slices], DATASET_NAME_1, 1, False,
                                     StorageInterface.ROOT_NODE_PATH)

            # Now try to read file
            read_data = self.storage.get_data(DATASET_NAME_2, None, StorageInterface.ROOT_NODE_PATH, False, True)
            self._assert_arrays_are_equal(self.test_2D_array, read_data)

    def test_add_metadata_non_tvb_specific(self):
        """
        This method checks metadata add for root or a dataset
        """
        # Create a dataset 
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        # Now add meta info to it.
        self.storage.set_metadata(META_DICT, DATASET_NAME_1, False, StorageInterface.ROOT_NODE_PATH)
        read_meta_value = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_value[META_KEY]

        # Now we'll test adding metadata on root node
        self.storage.set_metadata(META_DICT, '', False, StorageInterface.ROOT_NODE_PATH)
        read_meta_value = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_value[META_KEY]

    def test_delete_metadata_non_tvb_specific(self):
        """
        Test deletion of metadata for a dataset or root node.
        """
        # Create a dataset 
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        # Now add meta info to it.
        self.storage.set_metadata(META_DICT, DATASET_NAME_1, False, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_data[META_KEY]
        self.storage.remove_metadata(META_KEY, DATASET_NAME_1, False, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata(DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)
        assert 0 == len(read_meta_data)

        # Now we'll test removal of metadata on root node
        self.storage.set_metadata(META_DICT, '', False, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert META_VALUE == read_meta_data[META_KEY]
        self.storage.remove_metadata(META_KEY, '', False, StorageInterface.ROOT_NODE_PATH)
        read_meta_data = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        assert 1 == len(read_meta_data)

    def test_data_version(self):
        """
        Test if data version meta data is assigned when H5 file is created
        """
        self.storage.store_data(self.test_2D_array, DATASET_NAME_1, StorageInterface.ROOT_NODE_PATH)

        # Now read meta data for root node
        read_data = self.storage.get_metadata('', StorageInterface.ROOT_NODE_PATH)
        self._assert_arrays_are_equal(TvbProfile.current.version.DATA_VERSION,
                                      read_data[TvbProfile.current.version.DATA_VERSION_ATTRIBUTE])



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
This module is an interface for the tvb storage module.
All calls to methods from this module must be done through this class.

.. moduleauthor:: Robert Vincze <robert.vincze@codemart.ro>
"""

import os
import uuid
from datetime import datetime

from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.storage.h5.encryption.data_encryption_handler import FoldersQueueConsumer, \
    encryption_handler
from tvb.storage.h5.encryption.encryption_handler import EncryptionHandler
from tvb.storage.h5.encryption.import_export_encryption_handler import ImportExportEncryptionHandler
from tvb.storage.h5.file.exceptions import RenameWhileSyncEncryptingException, FileStructureException
from tvb.storage.h5.file.files_helper import FilesHelper, TvbZip
from tvb.storage.h5.file.hdf5_storage_manager import HDF5StorageManager
from tvb.storage.h5.file.xml_metadata_handlers import XMLReader, XMLWriter


class StorageInterface:
    TEMP_FOLDER = "TEMP"
    IMAGES_FOLDER = "IMAGES"
    PROJECTS_FOLDER = "PROJECTS"

    ROOT_NODE_PATH = "/"
    TVB_STORAGE_FILE_EXTENSION = ".h5"
    TVB_ZIP_FILE_EXTENSION = ".zip"
    TVB_XML_FILE_EXTENSION = ".xml"

    TVB_PROJECT_FILE = "Project" + TVB_XML_FILE_EXTENSION
    FILE_NAME_STRUCTURE = '{}_{}.h5'
    OPERATION_FOLDER_PREFIX = "Operation_"

    EXPORTED_SIMULATION_NAME = "exported_simulation"
    EXPORTED_SIMULATION_DTS_DIR = "datatypes"

    EXPORT_FOLDER_NAME = "EXPORT_TMP"
    EXPORT_FOLDER = os.path.join(TvbProfile.current.TVB_STORAGE, EXPORT_FOLDER_NAME)

    logger = get_logger(__name__)

    def __init__(self):
        self.files_helper = FilesHelper()
        self.data_encryption_handler = encryption_handler
        self.folders_queue_consumer = FoldersQueueConsumer()
        self.import_export_encryption_handler = ImportExportEncryptionHandler()

        # object attributes which have parameters in their constructor will be lazily instantiated
        self.tvb_zip = None
        self.xml_reader = None
        self.xml_writer = None

    # FilesHelper methods start here #

    def check_created(self, path=TvbProfile.current.TVB_STORAGE):
        self.files_helper.check_created(path)

    @staticmethod
    def get_projects_folder():
        return FilesHelper.get_projects_folder()

    def get_project_folder(self, project_name, *sub_folders):
        return self.files_helper.get_project_folder(project_name, *sub_folders)

    def get_temp_folder(self, project_name):
        return self.files_helper.get_project_folder(project_name, self.TEMP_FOLDER)

    def get_project_meta_file_path(self, project_name):
        return self.files_helper.get_project_meta_file_path(project_name)

    def read_project_metadata(self, project_path):
        return FilesHelper.read_project_metadata(project_path, self.TVB_PROJECT_FILE)

    def write_project_metadata_from_dict(self, project_path, meta_entity):
        FilesHelper.write_project_metadata_from_dict(project_path, meta_entity, self.TVB_PROJECT_FILE)

    def write_project_metadata(self, meta_dictionary):
        self.files_helper.write_project_metadata(meta_dictionary, self.TVB_PROJECT_FILE)

    def remove_operation_data(self, project_name, operation_id):
        self.files_helper.remove_operation_data(project_name, operation_id)

    def get_images_folder(self, project_name):
        return self.files_helper.get_images_folder(project_name, self.IMAGES_FOLDER)

    def write_image_metadata(self, figure, meta_entity):
        self.files_helper.write_image_metadata(figure, meta_entity, self.IMAGES_FOLDER)

    def remove_figure(self, figure):
        self.files_helper.remove_figure(figure, self.IMAGES_FOLDER)
        self.push_folder_to_sync(figure.project.name)

    def get_allen_mouse_cache_folder(self, project_name):
        return self.files_helper.get_allen_mouse_cache_folder(project_name)

    def get_tumor_dataset_folder(self):
        return self.files_helper.get_tumor_dataset_folder()

    @staticmethod
    def copy_file(source, dest, dest_postfix=None, buffer_size=1024 * 1024):
        FilesHelper.copy_file(source, dest, dest_postfix, buffer_size)

    @staticmethod
    def remove_files(file_list, ignore_exception=False):
        FilesHelper.remove_files(file_list, ignore_exception)

    @staticmethod
    def remove_folder(folder_path, ignore_errors=False):
        FilesHelper.remove_folder(folder_path, ignore_errors)

    @staticmethod
    def compute_size_on_disk(file_path):
        return FilesHelper.compute_size_on_disk(file_path)

    @staticmethod
    def compute_recursive_h5_disk_usage(start_path='.'):
        return FilesHelper.compute_recursive_h5_disk_usage(start_path)

    # TvbZip methods start here #

    def write_zip_folder(self, dest_path, folder, linked_paths=None, op=None, exclude=[]):
        self.tvb_zip = TvbZip(dest_path, "w")
        self.tvb_zip.write_zip_folder(folder, exclude)

        self.logger.debug("Done exporting files, now we will export linked DTs")

        if linked_paths is not None and op is not None:
            self.__export_datatypes(linked_paths, op)

        self.tvb_zip.close()

    def unpack_zip(self, uploaded_zip, folder_path):
        self.tvb_zip = TvbZip(uploaded_zip, "r")
        unpacked_folder = self.tvb_zip.unpack_zip(folder_path)
        self.tvb_zip.close()
        return unpacked_folder

    def get_filenames_in_zip(self, dest_path, mode="r"):
        self.tvb_zip = TvbZip(dest_path, mode)
        name_list = self.tvb_zip.namelist()
        self.tvb_zip.close()
        return name_list

    def open_tvb_zip(self, dest_path, name, mode="r"):
        self.tvb_zip = TvbZip(dest_path, mode)
        file = self.tvb_zip.open(name)
        self.tvb_zip.close()
        return file

    # Return a HDF5StorageManager object to call the methods from there #

    @staticmethod
    def get_storage_manager(file_full_path):
        return HDF5StorageManager(file_full_path)

    # XML methods start here #

    def read_metadata_from_xml(self, xml_path):
        self.xml_reader = XMLReader(xml_path)
        return self.xml_reader.read_metadata_from_xml()

    def write_metadata_in_xml(self, entity, final_path):
        self.xml_writer = XMLWriter(entity)
        return self.xml_writer.write_metadata_in_xml(final_path)

    # Method for preparing encryption
    def prepare_encryption(self, project_name):
        # Generate path to public_key
        temp_folder = self.get_temp_folder(project_name)
        public_key_file_name = "public_key_" + uuid.uuid4().hex + ".pem"
        public_key_file_path = os.path.join(temp_folder, public_key_file_name)

        # Generate a random password for the files
        password = EncryptionHandler.generate_random_password()

        return public_key_file_path, password

    # Return an EncryptionHandler object to call the methods from there #
    @staticmethod
    def get_encryption_handler(dir_gid):
        return EncryptionHandler(dir_gid)

    # Data Encryption Handler methods start here #

    def inc_project_usage_count(self, folder):
        return self.data_encryption_handler.inc_project_usage_count(folder)

    def inc_running_op_count(self, folder):
        return self.data_encryption_handler.inc_running_op_count(folder)

    def dec_running_op_count(self, folder):
        return self.data_encryption_handler.dec_running_op_count(folder)

    def check_and_delete(self, folder):
        self.data_encryption_handler.dec_running_op_count(folder)
        return self.data_encryption_handler.check_and_delete(folder)

    def sync_folders(self, folder):
        self.data_encryption_handler.sync_folders(folder)

    def set_project_active(self, project, linked_dt=None):
        self.data_encryption_handler.set_project_active(project, linked_dt)

    def set_project_inactive(self, project):
        self.data_encryption_handler.set_project_inactive(project.name)

    def push_folder_to_sync(self, project_name):
        project_folder = self.get_project_folder(project_name)
        self.data_encryption_handler.push_folder_to_sync(project_folder)

    @staticmethod
    def encryption_enabled():
        return encryption_handler.encryption_enabled()

    @staticmethod
    def app_encryption_handler():
        return encryption_handler.app_encryption_handler()

    def startup_cleanup(self):
        return self.data_encryption_handler.startup_cleanup()

    # Folders Queue Consumer methods start here #

    def run(self):
        return self.folders_queue_consumer.run()

    def start(self):
        self.folders_queue_consumer.start()

    def mark_stop(self):
        self.folders_queue_consumer.mark_stop()

    def join(self):
        self.folders_queue_consumer.join()

    # Generic methods start here

    def get_filename(self, class_name, gid):
        return self.FILE_NAME_STRUCTURE.format(class_name, gid.hex)

    def path_for(self, op_id, h5_file_class, gid, project_name, dt_class):
        operation_dir = self.files_helper.get_project_folder(project_name, str(op_id))
        return self.path_by_dir(operation_dir, h5_file_class, gid, dt_class)

    def path_by_dir(self, base_dir, h5_file_class, gid, dt_class):
        if isinstance(gid, str):
            gid = uuid.UUID(gid)
        fname = self.get_filename(dt_class or h5_file_class.file_name_base(), gid)
        return os.path.join(base_dir, fname)

    def ends_with_tvb_file_extension(self, file):
        return file.endswith(self.TVB_XML_FILE_EXTENSION)

    def ends_with_tvb_storage_file_extension(self, file):
        return file.endswith(self.TVB_STORAGE_FILE_EXTENSION)

    def is_in_usage(self, project_folder):
        return self.data_encryption_handler.is_in_usage(project_folder)

    def rename_project(self, current_proj_name, new_name):
        project_folder = self.get_project_folder(current_proj_name)
        if self.encryption_enabled() and not self.data_encryption_handler.is_in_usage(project_folder):
            raise RenameWhileSyncEncryptingException(
                "A project can not be renamed while sync encryption operations are running")
        self.files_helper.rename_project_structure(current_proj_name, new_name)
        encrypted_path = self.data_encryption_handler.compute_encrypted_folder_path(project_folder)
        if os.path.exists(encrypted_path):
            new_encrypted_path = self.data_encryption_handler.compute_encrypted_folder_path(
                self.get_project_folder(new_name))
            os.rename(encrypted_path, new_encrypted_path)

    def remove_project(self, project, sync_for_encryption=False):
        project_folder = self.get_project_folder(project.name)
        if sync_for_encryption:
            self.sync_folders(project_folder)
        try:
            self.remove_folder(project_folder)
            self.logger.debug("Project folders were removed for " + project.name)
        except OSError:
            self.logger.exception("A problem occurred while removing folder.")
            raise FileStructureException("Permission denied. Make sure you have write access on TVB folder!")

        encrypted_path = self.data_encryption_handler.compute_encrypted_folder_path(project_folder)
        FilesHelper.remove_files([encrypted_path, self.data_encryption_handler.project_key_path(project.id)], True)

    def move_datatype_with_sync(self, to_project, to_project_path, new_op_id, path_list):
        self.set_project_active(to_project)
        self.sync_folders(to_project_path)

        for path in path_list:
            self.files_helper.move_datatype(to_project.name, str(new_op_id), path)

        self.sync_folders(to_project_path)
        self.set_project_inactive(to_project)

    # Return an ImportExportEncryptionHandler object  to call the methods from there #

    @staticmethod
    def get_import_export_encryption_handler():
        return ImportExportEncryptionHandler()

    # Exporting related methods start here

    def __export_datatypes(self, paths, operation):
        op_folder = self.get_project_folder(operation.project.name, operation.id)
        op_folder_name = os.path.basename(op_folder)

        # add linked datatypes to archive in the import operation
        for pth in paths:
            zip_pth = op_folder_name + '/' + os.path.basename(pth)
            self.tvb_zip.write(pth, zip_pth)

        # remove these files, since we only want them in export archive
        self.remove_folder(op_folder)

    def __build_data_export_folder(self, data, export_folder):
        now = datetime.now()
        date_str = "%d-%d-%d_%d-%d-%d_%d" % (now.year, now.month, now.day, now.hour,
                                             now.minute, now.second, now.microsecond)
        tmp_str = date_str + "@" + data.gid
        data_export_folder = os.path.join(export_folder, tmp_str)
        self.check_created(data_export_folder)

        return data_export_folder

    def export_project(self, project, folders_to_exclude, linked_paths, op):
        """
        This method is used to export a project as a ZIP file.
        :param project: project to be exported.
        :param folders_to_exclude: a list of paths to folders inside of a project folder which should not be exported.
        :param linked_paths: a list of links to datatypes for the project to be exported
        :param op: operation for links to exported datatypes (if any)
        """

        project_folder = self.get_project_folder(project.name)
        folders_to_exclude.append("TEMP")

        # Compute path and name of the zip file
        now = datetime.now()
        date_str = now.strftime("%Y-%m-%d_%H-%M")
        zip_file_name = "%s_%s.%s" % (date_str, project.name, self.TVB_ZIP_FILE_EXTENSION)

        export_folder = self.__build_data_export_folder(project, self.EXPORT_FOLDER)
        result_path = os.path.join(export_folder, zip_file_name)

        # Pack project [filtered] content into a ZIP file:
        self.logger.debug("Done preparing, now we will write the folder.")
        self.logger.debug(project_folder)
        self.write_zip_folder(result_path, project_folder, linked_paths, op, folders_to_exclude)

        # Make sure the Project.xml file gets copied:
        self.logger.debug("Done, closing")

        return result_path

    def export_simulator_configuration(self, burst, all_view_model_paths, all_datatype_paths, zip_filename):
        """
        This method is used to export a simulator configuration as a ZIP file
        :param burst: BurstConfiguration of the simulation to be exported
        :param all_view_model_paths: a list of paths to all view model files of the simulation
        :param all_datatype_paths: a list of paths to all datatype files of the simulation
        :param zip_filename: name of the file to be exported
        """

        tmp_export_folder = self.__build_data_export_folder(burst, self.EXPORT_FOLDER)
        tmp_sim_folder = os.path.join(tmp_export_folder, self.EXPORTED_SIMULATION_NAME)

        if not os.path.exists(tmp_sim_folder):
            os.makedirs(tmp_sim_folder)

        for vm_path in all_view_model_paths:
            dest = os.path.join(tmp_sim_folder, os.path.basename(vm_path))
            self.copy_file(vm_path, dest)

        for dt_path in all_datatype_paths:
            dest = os.path.join(tmp_sim_folder, self.EXPORTED_SIMULATION_DTS_DIR, os.path.basename(dt_path))
            self.copy_file(dt_path, dest)

        result_path = os.path.join(tmp_export_folder, zip_filename)
        self.write_zip_folder(result_path, tmp_sim_folder)
        self.remove_folder(tmp_sim_folder)

        return result_path

    def __copy_datatypes(self, dt_path_list, data, password):
        export_folder = self.__build_data_export_folder(data, self.EXPORT_FOLDER)

        for dt_path in dt_path_list:
            file_destination = os.path.join(export_folder, os.path.basename(dt_path))
            if not os.path.exists(file_destination):
                self.copy_file(dt_path, file_destination)
                self.get_storage_manager(file_destination).remove_metadata('parent_burst', check_existence=True)

                if password is not None:
                    self.import_export_encryption_handler.encrypt_data_at_export(file_destination, password)
                    os.remove(file_destination)

        return export_folder

    def export_datatypes(self, dt_path_list, data, download_file_name, public_key_path=None, password=None):
        """
        This method is used to export a list of datatypes as a ZIP file.
        :param dt_path_list: a list of paths to be exported (there are more than one when exporting with links)
        :param data: data to be exported
        :param download_file_name: name of the zip file to be downloaded
        :param public_key_path: path to public key that will be used for encrypting the password by TVB
        :param password: password used for encrypting the files before exporting
        """

        export_folder = self.__copy_datatypes(dt_path_list, data, password)

        if len(dt_path_list) == 1 and password is None:
            return os.path.join(export_folder, os.path.basename(dt_path_list[0]))

        if password is not None:
            download_file_name = download_file_name.replace('.h5', '.zip')
            self.import_export_encryption_handler.encrypt_and_save_password(public_key_path, password, export_folder)

        export_data_zip_path = os.path.join(os.path.dirname(export_folder), download_file_name)
        self.write_zip_folder(export_data_zip_path, export_folder)
        return export_data_zip_path

    def export_datatypes_structure(self, op_file_dict, data, download_file_name, public_key_path, password,
                                   links_tuple_for_copy=None):
        """
        This method is used to export a list of datatypes as a ZIP file, while preserving the folder structure
        (eg: operation folders). It is only used during normal tvb exporting for datatype groups.
        :param op_file_dict: a dictionary where keys are operation folders and the values are lists of files inside
            that operation folder
        :param data: data to be exported
        :param download_file_name: name of the ZIP file to be exported
        :param public_key_path: path to public key that will be used for encrypting the password by TVB
        :param password: password used for encrypting the files before exporting
        :param links_tuple_for_copy: a tuple containing two elements: a list of paths to be copied and the first
         datatype of the group

        """
        if links_tuple_for_copy is not None:
            export_folder = self.__copy_datatypes(links_tuple_for_copy[0], links_tuple_for_copy[1], password)
        else:
            export_folder = self.__build_data_export_folder(data, self.EXPORT_FOLDER)

        for op_folder, files in op_file_dict.items():
            tmp_op_folder_path = os.path.join(export_folder, os.path.basename(op_folder))
            for file in files:
                dest_path = os.path.join(tmp_op_folder_path, os.path.basename(file))

                if not os.path.exists(dest_path):
                    self.copy_file(file, os.path.join(tmp_op_folder_path, os.path.basename(file)))
                    self.get_storage_manager(dest_path).remove_metadata('parent_burst', check_existence=True)

                    if password is not None:
                        self.import_export_encryption_handler.encrypt_data_at_export(dest_path, password)
                        os.remove(dest_path)

        dest_path = os.path.join(os.path.dirname(export_folder), download_file_name)
        if password is not None:
            self.import_export_encryption_handler.encrypt_and_save_password(public_key_path, password, export_folder)
        self.write_zip_folder(dest_path, export_folder)

        return dest_path

    def export_datatype_from_rest_server(self, dt, data, download_file_name, public_key_path):
        password = EncryptionHandler.generate_random_password()
        dest_path = self.export_datatypes([dt], data, download_file_name, public_key_path, password)

        return dest_path

    def import_datatype_to_rest_client(self, file_path, temp_folder, private_key_path):
        import_export_encryption_handler = self.get_import_export_encryption_handler()
        result = self.unpack_zip(file_path, temp_folder)
        encrypted_password_path = import_export_encryption_handler.extract_encrypted_password_from_list(result)

        decrypted_file_path = import_export_encryption_handler.decrypt_content(encrypted_password_path,
                                                                               result, private_key_path)[0]
        os.remove(file_path)
        return decrypted_file_path



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
This module contains the necessary methods for encryption and decryption at import/export.

.. moduleauthor:: Robert Vincze <robert.vincze@codemart.ro>
"""

import os
import pyAesCrypt

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from tvb.basic.profile import TvbProfile


class ImportExportEncryptionHandler:
    ENCRYPTED_DATA_SUFFIX = '_encrypted'
    ENCRYPTED_PASSWORD_NAME = 'encrypted_password.pem'
    DECRYPTED_DATA_SUFFIX = '_decrypted'
    PUBLIC_KEY_NAME = 'public_key.pem'
    PRIVATE_KEY_NAME = 'private_key.pem'

    def generate_public_private_key_pair(self, key_path):
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

        public_key = private_key.public_key()
        # The public key will be uploaded to TVB before exporting and the private key will be used for
        # decrypting the data locally

        # Step 2. Convert public key to bytes and save it so ABCUploader can use it
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.PKCS1,
        )

        public_key_path = os.path.join(key_path, self.PUBLIC_KEY_NAME)
        with open(public_key_path, 'wb') as f:
            f.write(pem)

        # Step 3. Convert private key to bytes and save it so it can be used later for decryption
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )

        private_key_path = os.path.join(key_path, self.PRIVATE_KEY_NAME)
        with open(private_key_path, 'wb') as f:
            f.write(pem)

    def add_encrypted_suffix(self, name):
        return name.replace('.zip', self.ENCRYPTED_DATA_SUFFIX + '.zip')

    def get_path_to_encrypt(self, input_path):
        start_extension = input_path.rfind('.')
        path_to_encrypt = input_path[:start_extension]
        extension = input_path[start_extension:]

        return path_to_encrypt + self.ENCRYPTED_DATA_SUFFIX + extension

    @staticmethod
    def load_public_key(public_key_path):
        with open(public_key_path, "rb") as key_file:
            public_key = serialization.load_pem_public_key(key_file.read(), backend=default_backend())

        return public_key

    @staticmethod
    def encrypt_password(public_key, symmetric_key):

        encrypted_symmetric_key = public_key.encrypt(
            symmetric_key,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        return encrypted_symmetric_key

    def save_encrypted_password(self, encrypted_password, path_to_encrypted_password):
        save_password_path = os.path.join(path_to_encrypted_password, self.ENCRYPTED_PASSWORD_NAME)
        with open(os.path.join(save_password_path), 'wb') as f:
            f.write(encrypted_password)

        return save_password_path

    def encrypt_and_save_password(self, public_key_path, password, path_to_encrypted_password):
        public_key = self.load_public_key(public_key_path)
        password_bytes = str.encode(password)
        encrypted_password = self.encrypt_password(public_key, password_bytes)
        self.save_encrypted_password(encrypted_password, path_to_encrypted_password)

    def encrypt_data_at_export(self, file_path, password):
        # Encrypt the file(s) using the generated password
        buffer_size = TvbProfile.current.hpc.CRYPT_BUFFER_SIZE

        encrypted_file_path = self.get_path_to_encrypt(file_path)
        pyAesCrypt.encryptFile(file_path, encrypted_file_path, password, buffer_size)

        return encrypted_file_path

    def decrypt_content(self, encrypted_aes_key_path, upload_paths, private_key_path):

        # Get the encrypted password
        with open(encrypted_aes_key_path, 'rb') as f:
            encrypted_password = f.read()

        # Read the private key
        with open(private_key_path, "rb") as key_file:
            private_key = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
            )

        # Decrypt the password using the private key
        decrypted_password = private_key.decrypt(
            encrypted_password,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        decrypted_password = decrypted_password.decode()

        # Get path to decrypted file
        decrypted_paths = []

        for path in upload_paths:
            decrypted_download_path = path.replace(self.ENCRYPTED_DATA_SUFFIX,
                                                   self.DECRYPTED_DATA_SUFFIX)
            decrypted_paths.append(decrypted_download_path)

            # Use the decrypted password to decrypt the message
            pyAesCrypt.decryptFile(path, decrypted_download_path, decrypted_password,
                                   TvbProfile.current.hpc.CRYPT_BUFFER_SIZE)
        return decrypted_paths

    def extract_encrypted_password_from_list(self, file_list):
        password = None
        password_idx = None

        for idx, file in enumerate(file_list):
            if self.ENCRYPTED_PASSWORD_NAME in file:
                password = file
                password_idx = idx

        del file_list[password_idx]
        return password



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Persistence of data in HDF5 format.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
"""

import copy
import os
import threading
from datetime import datetime
import h5py as hdf5
import numpy as numpy

from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.storage.h5 import utils
from tvb.storage.h5.encryption.data_encryption_handler import encryption_handler
from tvb.storage.h5.file.exceptions import MissingDataSetException, IncompatibleFileManagerException, \
    FileStructureException, MissingDataFileException
from tvb.storage.h5.file.files_helper import FilesHelper

# Create logger for this module
LOG = get_logger(__name__)

LOCK_OPEN_FILE = threading.Lock()
BUFFER_SIZE = 300


class HDF5StorageManager(object):
    """
    This class is responsible for saving / loading data in HDF5 file / format.
    """
    __file_title_ = "TVB data file"
    __storage_full_name = None
    __hfd5_file = None

    TVB_ATTRIBUTE_PREFIX = "TVB_"
    ROOT_NODE_PATH = "/"
    BOOL_VALUE_PREFIX = "bool:"
    DATETIME_VALUE_PREFIX = "datetime:"
    DATE_TIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
    LOCKS = {}

    def __init__(self, storage_full_name):
        """
        Creates a new storage manager instance.
        :param buffer_size: the size in Bytes of the amount of data that will be buffered before writing to file.
        """
        if storage_full_name is None:
            raise FileStructureException("Please provide the file name where to store data")

        self.__storage_full_name = storage_full_name
        self.__buffer_array = None
        self.data_buffers = {}
        self.data_encryption_handler = encryption_handler

    def is_valid_tvb_file(self):
        """
        This method checks if specified file exists and if it has correct HDF5 format
        :returns: True is file exists and has HDF5 format. False otherwise.
        """
        try:
            return os.path.exists(self.__storage_full_name) and hdf5.is_hdf5(self.__storage_full_name)
        except RuntimeError:
            return False

    def store_data(self, data_list, dataset_name='', where=ROOT_NODE_PATH):
        """
        This method stores provided data list into a data set in the H5 file.

        :param dataset_name: Name of the data set where to store data
        :param data_list: Data to be stored
        :param where: represents the path where to store our dataset (e.g. /data/info)
        """
        data_to_store = self._check_data(data_list)

        try:
            LOG.debug("Saving data into data set: %s" % dataset_name)
            # Open file in append mode ('a') to allow adding multiple data sets in the same file
            hdf5_file = self._open_h5_file()

            full_dataset_name = where + dataset_name
            if full_dataset_name not in hdf5_file:
                hdf5_file.create_dataset(full_dataset_name, data=data_to_store)

            elif hdf5_file[full_dataset_name].shape == data_to_store.shape:
                hdf5_file[full_dataset_name][...] = data_to_store[...]

            else:
                raise IncompatibleFileManagerException("Cannot update existing H5 DataSet %s with a different shape. "
                                                       "Try defining it as chunked!" % full_dataset_name)

        finally:
            # Now close file
            self.close_file()
            self.data_encryption_handler.push_folder_to_sync(FilesHelper.get_project_folder_from_h5(
                self.__storage_full_name))

    def append_data(self, data_list, dataset_name='', grow_dimension=-1, close_file=True, where=ROOT_NODE_PATH):
        """
        This method appends data to an existing data set. If the data set does not exists, create it first.

        :param dataset_name: Name of the data set where to store data
        :param data_list: Data to be stored / appended
        :param grow_dimension: The dimension to be used to grow stored array. By default will grow on the LAST dimension
        :param close_file: Specify if the file should be closed automatically after write operation. If not,
            you have to close file by calling method close_file()
        :param where: represents the path where to store our dataset (e.g. /data/info)

        """
        data_to_store = self._check_data(data_list)
        data_buffer = self.data_buffers.get(where + dataset_name, None)

        if data_buffer is None:
            hdf5_file = self._open_h5_file()
            datapath = where + dataset_name
            if datapath in hdf5_file:
                dataset = hdf5_file[datapath]
                self.data_buffers[datapath] = HDF5StorageManager.H5pyStorageBuffer(dataset,
                                                                                   buffered_data=data_to_store,
                                                                                   grow_dimension=grow_dimension)
            else:
                data_shape_list = list(data_to_store.shape)
                data_shape_list[grow_dimension] = None
                data_shape = tuple(data_shape_list)
                dataset = hdf5_file.create_dataset(where + dataset_name, data=data_to_store, shape=data_to_store.shape,
                                                   dtype=data_to_store.dtype, maxshape=data_shape)
                self.data_buffers[datapath] = HDF5StorageManager.H5pyStorageBuffer(dataset,
                                                                                   buffered_data=None,
                                                                                   grow_dimension=grow_dimension)
        else:
            if not data_buffer.buffer_data(data_to_store):
                data_buffer.flush_buffered_data()
        if close_file:
            self.close_file()
        self.data_encryption_handler.push_folder_to_sync(
            FilesHelper.get_project_folder_from_h5(self.__storage_full_name))

    def remove_data(self, dataset_name='', where=ROOT_NODE_PATH):
        """
        Deleting a data set from H5 file.

        :param dataset_name:name of the data set to be deleted
        :param where: represents the path where dataset is stored (e.g. /data/info)

        """
        LOG.debug("Removing data set: %s" % dataset_name)
        try:
            # Open file in append mode ('a') to allow data remove
            hdf5_file = self._open_h5_file()
            del hdf5_file[where + dataset_name]

        except KeyError:
            LOG.warning("Trying to delete data set: %s but current file does not contain it." % dataset_name)
            raise FileStructureException("Could not locate dataset: %s" % dataset_name)
        finally:
            self.close_file()
            self.data_encryption_handler.push_folder_to_sync(
                FilesHelper.get_project_folder_from_h5(self.__storage_full_name))

    def get_data(self, dataset_name='', data_slice=None, where=ROOT_NODE_PATH, ignore_errors=False, close_file=True):
        """
        This method reads data from the given data set based on the slice specification

        :param close_file: Automatically close after reading the current field
        :param ignore_errors: return None in case of error, or throw exception
        :param dataset_name: Name of the data set from where to read data
        :param data_slice: Specify how to retrieve data from array {e.g (slice(1,10,1),slice(1,6,2)) }
        :param where: represents the path where dataset is stored (e.g. /data/info)
        :returns: a numpy.ndarray containing filtered data

        """
        LOG.debug("Reading data from data set: %s" % dataset_name)

        data_path = where + dataset_name
        try:
            # Open file to read data
            hdf5_file = self._open_h5_file('r')
            if data_path in hdf5_file:
                data_array = hdf5_file[data_path]
                # Now read data
                if data_slice is None:
                    result = data_array[()]
                    if isinstance(result, hdf5.Empty):
                        return numpy.empty([])
                    return result
                else:
                    return data_array[data_slice]
            else:
                if not ignore_errors:
                    LOG.error("Trying to read data from a missing data set: %s" % dataset_name)
                    raise MissingDataSetException("Could not locate dataset: %s" % dataset_name)
                else:
                    return None
        finally:
            if close_file:
                self.close_file()

    def get_data_shape(self, dataset_name='', where=ROOT_NODE_PATH):
        """
        This method reads data-size from the given data set

        :param dataset_name: Name of the data set from where to read data
        :param where: represents the path where dataset is stored (e.g. /data/info)
        :returns: a tuple containing data size
        
        """
        LOG.debug("Reading data from data set: %s" % dataset_name)
        try:
            # Open file to read data
            hdf5_file = self._open_h5_file('r')
            data_array = hdf5_file[where + dataset_name]
            return data_array.shape
        except KeyError:
            LOG.debug("Trying to read data from a missing data set: %s" % dataset_name)
            raise MissingDataSetException("Could not locate dataset: %s" % dataset_name)

        finally:
            self.close_file()

    def set_metadata(self, meta_dictionary, dataset_name='', tvb_specific_metadata=True, where=ROOT_NODE_PATH):
        """
        Set meta-data information for root node or for a given data set.

        :param meta_dictionary: dictionary containing meta info to be stored on node
        :param dataset_name: name of the dataset where to assign metadata. If None, metadata is assigned to ROOT node.
        :param tvb_specific_metadata: specify if the provided metadata is TVB specific (All keys will have a TVB prefix)
        :param where: represents the path where dataset is stored (e.g. /data/info)

        """
        LOG.debug("Setting metadata on node: %s" % dataset_name)

        # Open file to read data
        hdf5_file = self._open_h5_file()
        try:
            node = hdf5_file[where + dataset_name]
        except KeyError:
            LOG.debug("Trying to set metadata on a missing data set: %s" % dataset_name)
            node = hdf5_file.create_dataset(where + dataset_name, (1,))

        try:
            # Now set meta-data
            for meta_key in meta_dictionary:
                key_to_store = meta_key
                if tvb_specific_metadata:
                    key_to_store = self.TVB_ATTRIBUTE_PREFIX + meta_key

                processed_value = self._serialize_value(meta_dictionary[meta_key])
                node.attrs[key_to_store] = processed_value
        finally:
            self.close_file()
            self.data_encryption_handler.push_folder_to_sync(
                FilesHelper.get_project_folder_from_h5(self.__storage_full_name))

    @staticmethod
    def serialize_bool(value):
        return HDF5StorageManager.BOOL_VALUE_PREFIX + str(value)

    def _serialize_value(self, value):
        """
        This method takes a value which will be stored as metadata and
        apply some transformation if necessary

        :param value: value which is planned to be stored
        :returns:  value to be stored

        """
        if value is None:
            return ''
        # Transform boolean to string and prefix it
        if isinstance(value, bool):
            return self.serialize_bool(value)
        # Transform date to string and append prefix
        elif isinstance(value, datetime):
            return self.DATETIME_VALUE_PREFIX + utils.date2string(value, date_format=self.DATE_TIME_FORMAT)
        else:
            return value

    def remove_metadata(self, meta_key, dataset_name='', tvb_specific_metadata=True, where=ROOT_NODE_PATH,
                        check_existence=False):
        """
        Remove meta-data information for root node or for a given data set.

        :param meta_key: name of the metadata attribute to be removed
        :param dataset_name: name of the dataset from where to delete metadata.
            If None, metadata will be removed from ROOT node.
        :param tvb_specific_metadata: specify if the provided metadata is specific to TVB (keys will have a TVB prefix).
        :param where: represents the path where dataset is stored (e.g. /data/info)

        """

        if check_existence and meta_key not in self.get_metadata():
            return

        LOG.debug("Deleting metadata: %s for dataset: %s" % (meta_key, dataset_name))
        try:
            # Open file to read data
            hdf5_file = self._open_h5_file()
            node = hdf5_file[where + dataset_name]

            # Now delete metadata
            key_to_remove = meta_key
            if tvb_specific_metadata:
                key_to_remove = self.TVB_ATTRIBUTE_PREFIX + meta_key
            del node.attrs[key_to_remove]
        except KeyError:
            LOG.error("Trying to delete metadata on a missing data set: %s" % dataset_name)
            raise FileStructureException("Could not locate dataset: %s" % dataset_name)
        except AttributeError:
            LOG.error("Trying to delete missing metadata %s" % meta_key)
            raise FileStructureException("There is no metadata named %s on this node" % meta_key)
        finally:
            self.close_file()
            self.data_encryption_handler.push_folder_to_sync(
                FilesHelper.get_project_folder_from_h5(self.__storage_full_name))

    def get_metadata(self, dataset_name='', where=ROOT_NODE_PATH):
        """
        Retrieve ALL meta-data information for root node or for a given data set.

        :param dataset_name: name of the dataset for which to read metadata. If None, read metadata from ROOT node.
        :param where: represents the path where dataset is stored (e.g. /data/info)
        :returns: a dictionary containing all metadata associated with the node

        """
        LOG.debug("Retrieving metadata for dataset: %s" % dataset_name)
        meta_key = ""
        try:
            # Open file to read data
            hdf5_file = self._open_h5_file('r')
            node = hdf5_file[where + dataset_name]
            # Now retrieve metadata values
            all_meta_data = {}

            for meta_key in node.attrs:
                new_key = meta_key
                if meta_key.startswith(self.TVB_ATTRIBUTE_PREFIX):
                    new_key = meta_key[len(self.TVB_ATTRIBUTE_PREFIX):]
                value = node.attrs[meta_key]
                all_meta_data[new_key] = self._deserialize_value(value)
            return all_meta_data

        except KeyError:
            msg = "Trying to read data from a missing data set: %s" % (where + dataset_name)
            LOG.warning(msg)
            raise MissingDataSetException(msg)
        except AttributeError:
            msg = "Trying to get value for missing metadata %s" % meta_key
            LOG.exception(msg)
            raise FileStructureException(msg)
        except Exception:

            msg = "Failed to read metadata from H5 file! %s" % self.__storage_full_name
            LOG.exception(msg)
            raise FileStructureException(msg)
        finally:
            self.close_file()

    def get_file_data_version(self, data_version, dataset_name='', where=ROOT_NODE_PATH):
        """
        Checks the data version for the current file.
        """
        if not os.path.exists(self.__storage_full_name):
            raise MissingDataFileException("File storage data not found at path %s" % (self.__storage_full_name,))

        if self.is_valid_tvb_file():
            metadata = self.get_metadata(dataset_name, where)
            if data_version in metadata:
                return metadata[data_version]
            else:
                raise IncompatibleFileManagerException("Could not find TVB specific data version attribute %s in file: "
                                                       "%s." % (data_version, self.__storage_full_name))
        raise IncompatibleFileManagerException("File %s is not a hdf5 format file. Are you using the correct "
                                               "manager for this file?" % (self.__storage_full_name,))

    def _deserialize_value(self, value):
        """
        This method takes value loaded from H5 file and transform it to TVB data.
        """
        if value is not None:
            if isinstance(value, numpy.string_):
                if len(value) == 0:
                    value = None
                else:
                    value = str(value)

            if isinstance(value, str):
                if value.startswith(self.BOOL_VALUE_PREFIX):
                    # Remove bool prefix and transform to bool
                    return utils.string2bool(value[len(self.BOOL_VALUE_PREFIX):])
                if value.startswith(self.DATETIME_VALUE_PREFIX):
                    # Remove datetime prefix and transform to datetime
                    return utils.string2date(value[len(self.DATETIME_VALUE_PREFIX):], date_format=self.DATE_TIME_FORMAT)

        return value

    def __aquire_lock(self):
        """
        Aquire a unique lock for each different file path on the system.
        """
        lock = self.LOCKS.get(self.__storage_full_name, None)
        if lock is None:
            lock = threading.Lock()
            self.LOCKS[self.__storage_full_name] = lock
        lock.acquire()

    def __release_lock(self):
        """
        Aquire a unique lock for each different file path on the system.
        """
        lock = self.LOCKS.get(self.__storage_full_name, None)
        if lock is None:
            raise Exception("Some lock was deleted without being released beforehand.")
        lock.release()

    def close_file(self):
        """
        The synchronization of open/close doesn't seem to be needed anymore for h5py in
        contrast to PyTables for concurrent reads. However since it shouldn't add that
        much overhead in most situation we'll leave it like this for now since in case
        of concurrent writes(metadata) this provides extra safety.
        """
        try:
            self.__aquire_lock()
            self.__close_file()
        finally:
            self.__release_lock()

    def _open_h5_file(self, mode='a'):
        """
        The synchronization of open/close doesn't seem to be needed anymore for h5py in
        contrast to PyTables for concurrent reads. However since it shouldn't add that
        much overhead in most situation we'll leave it like this for now since in case
        of concurrent writes(metadata) this provides extra safety.
        """
        try:
            self.__aquire_lock()
            file_obj = self.__open_h5_file(mode)
        finally:
            self.__release_lock()
        return file_obj

    def __close_file(self):
        """
        Close file used to store data.
        """
        hdf5_file = self.__hfd5_file

        # Try to close file only if it was opened before
        if hdf5_file is not None and hdf5_file.id.valid:
            LOG.debug("Closing file: %s" % self.__storage_full_name)
            try:
                for h5py_buffer in self.data_buffers.values():
                    h5py_buffer.flush_buffered_data()
                self.data_buffers = {}
                hdf5_file.close()
            except Exception as excep:
                # Do nothing is this situation.
                # The file is correctly closed, but the list of open files on HDF5 is not updated in a synch manner.
                # del _open_files[filename] might throw KeyError
                LOG.exception(excep)
            if not hdf5_file.id.valid:
                self.__hfd5_file = None

    # -------------- Private methods  --------------
    def __open_h5_file(self, mode='a'):
        """
        Open file for reading, writing or append.

        :param mode: Mode to open file (possible values are w / r / a).
                    Default value is 'a', to allow adding multiple data to the same file.
        :returns: returns the file which stores data in HDF5 format opened for read / write according to mode param

        """
        if self.__storage_full_name is None:
            raise FileStructureException("Invalid storage file. Please provide a valid path.")
        try:
            # Check if file is still open from previous writes.
            if self.__hfd5_file is None or not self.__hfd5_file.id.valid:
                file_exists = os.path.exists(self.__storage_full_name)

                # bug in some versions of hdf5 on windows prevent creating file with mode='a'
                if not file_exists and mode == 'a':
                    mode = 'w'

                LOG.debug("Opening file: %s in mode: %s" % (self.__storage_full_name, mode))
                self.__hfd5_file = hdf5.File(self.__storage_full_name, mode, libver='latest')

                # If this is the first time we access file, write data version
                if not file_exists:
                    os.chmod(self.__storage_full_name, TvbProfile.current.ACCESS_MODE_TVB_FILES)
                    attr_name = self.TVB_ATTRIBUTE_PREFIX + TvbProfile.current.version.DATA_VERSION_ATTRIBUTE
                    self.__hfd5_file['/'].attrs[attr_name] = TvbProfile.current.version.DATA_VERSION
        except (IOError, OSError) as err:
            LOG.exception("Could not open storage file.")
            raise FileStructureException("Could not open storage file. %s" % err)

        return self.__hfd5_file

    @staticmethod
    def _check_data(data_list):
        """
        Check if the data to be stores is in a good format. If not adapt it.
        """
        if data_list is None:
            raise FileStructureException("Could not store null data")

        if not (isinstance(data_list, list) or isinstance(data_list, numpy.ndarray)):
            raise FileStructureException("Invalid data type. Could not store data of type:" + str(type(data_list)))

        data_to_store = data_list
        if isinstance(data_to_store, list):
            data_to_store = numpy.array(data_list)
        if data_to_store.shape == ():
            data_to_store = hdf5.Empty("f")
        return data_to_store

    class H5pyStorageBuffer(object):
        """
        Helper class in order to buffer data for append operations, to limit the number of actual
        HDD I/O operations.
        """

        def __init__(self, h5py_dataset, buffered_data=None, grow_dimension=-1):
            self.buffered_data = buffered_data
            self.buffer_size = BUFFER_SIZE
            if h5py_dataset is None:
                raise MissingDataSetException("A H5pyStorageBuffer instance must have a h5py dataset for which the"
                                              "buffering is done. Please supply one to the 'h5py_dataset' parameter.")
            self.h5py_dataset = h5py_dataset
            self.grow_dimension = grow_dimension

        def buffer_data(self, data_list):
            """
            Add data_list to an internal buffer in order to improve performance for append_data type of operations.
            :returns: True if buffer is still fine, \
                      False if a flush is necessary since the buffer is full
            """
            if self.buffered_data is None:
                self.buffered_data = data_list
            else:
                self.buffered_data = self.__custom_numpy_append(self.buffered_data, data_list)
            if self.buffered_data.nbytes > self.buffer_size:
                return False
            else:
                return True

        def __custom_numpy_append(self, array1, array2):
            array_1_shape = numpy.array(array1.shape)
            array_2_shape = numpy.array(array2.shape)
            result_shape = copy.deepcopy(array_1_shape)
            result_shape[self.grow_dimension] += array_2_shape[self.grow_dimension]
            result_array = numpy.empty(shape=tuple(result_shape), dtype=array1.dtype)
            full_slice = slice(None, None, None)
            full_index = [full_slice for _ in array_1_shape]
            full_index[self.grow_dimension] = slice(0, array_1_shape[self.grow_dimension], None)
            result_array[tuple(full_index)] = array1
            full_index[self.grow_dimension] = slice(array_1_shape[self.grow_dimension],
                                                    result_shape[self.grow_dimension], None)
            result_array[tuple(full_index)] = array2
            return result_array

        def flush_buffered_data(self):
            """
            Append the data buffered so far to the input dataset using :param grow_dimension: as the dimension that
            will be expanded.
            """
            if self.buffered_data is not None:
                current_shape = self.h5py_dataset.shape
                new_shape = list(current_shape)
                new_shape[self.grow_dimension] += self.buffered_data.shape[self.grow_dimension]
                # Create the required slice to which the new data will be added.
                # For example if the 3nd dimension of a 4D datashape (74, 1, 100, 1)
                # we want to get the slice (:, :, 100:200, :) in order to add 100 new entries
                full_slice = slice(None, None, None)
                slice_to_add = slice(current_shape[self.grow_dimension], new_shape[self.grow_dimension], None)
                append2address = [full_slice for _ in new_shape]
                append2address[self.grow_dimension] = slice_to_add
                # Do the data reshape and copy the new data
                self.h5py_dataset.resize(tuple(new_shape))
                self.h5py_dataset[tuple(append2address)] = self.buffered_data
                self.buffered_data = None



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Exceptions for File Storage layer. 
   
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""
from tvb.basic.exceptions import TVBException


class FileStructureException(TVBException):
    """
    Exception to be thrown in case of a problem 
    related to File Structure Storage.
    """

    def __init__(self, message):
        TVBException.__init__(self, message)


class FileStorageException(TVBException):
    """
    Generic exception when storing in data in files.
    """

    def __init__(self, message):
        TVBException.__init__(self, message)


class MissingDataSetException(FileStorageException):
    """
    Exception when a dataset is accessed, but no written entry exists in HDF5 file for it.
    we will consider the attribute None.
    """

    def __init__(self, message):
        FileStorageException.__init__(self, message)


class MissingDataFileException(FileStorageException):
    """
    Exception when the file associated to some manager does not exist on disk for some reason.
    """

    def __init__(self, message):
        FileStorageException.__init__(self, message)


class FileVersioningException(TVBException):
    """
    A base exception class for all TVB file storage version conversion custom exceptions.
    """

    def __init__(self, message):
        TVBException.__init__(self, message)


class IncompatibleFileManagerException(FileVersioningException):
    """
    Exception that should be raised in case a file is handled by some filemanager which
    is incompatible with that version of TVB file storage.
    """

    def __init__(self, message):
        FileVersioningException.__init__(self, message)


class FileMigrationException(TVBException):
    """
    Exception to be thrown in case of an unexpected problem
    when migrating an H5 file to a newer version.
    """

    def __init__(self, message):
        TVBException.__init__(self, message)


class UnsupportedFileStorageException(TVBException):
    """
    Exception to be thrown in case of an unsupported file storage is chosen.
    Currently only H5 Storage is supported.
    """

    def __init__(self, message):
        super().__init__(message)


class RenameWhileSyncEncryptingException(TVBException):
    """
    Exception to be thrown in case a project is to be renamed during sync encryption.
    """

    def __init__(self, message):
        super().__init__(message)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from datetime import datetime

COMPLEX_TIME_FORMAT = '%Y-%m-%d,%H-%M-%S.%f'
# LESS_COMPLEX_TIME_FORMAT is also compatible with data exported from TVB 1.0.
# This is only used as a fallback in the string to date conversion.
LESS_COMPLEX_TIME_FORMAT = '%Y-%m-%d,%H-%M-%S'
SIMPLE_TIME_FORMAT = "%m-%d-%Y"


def string2date(string_input, complex_format=True, date_format=None):
    """Read date from string, after internal format"""
    if string_input == 'None':
        return None
    if date_format is not None:
        return datetime.strptime(string_input, date_format)
    if complex_format:
        try:
            return datetime.strptime(string_input, COMPLEX_TIME_FORMAT)
        except ValueError:
            # For backwards compatibility with TVB 1.0
            return datetime.strptime(string_input, LESS_COMPLEX_TIME_FORMAT)
    return datetime.strptime(string_input, SIMPLE_TIME_FORMAT)


def date2string(date_input, complex_format=True, date_format=None):
    """Convert date into string, after internal format"""
    if date_input is None:
        return "None"

    if date_format is not None:
        return date_input.strftime(date_format)

    if complex_format:
        return date_input.strftime(COMPLEX_TIME_FORMAT)
    return date_input.strftime(SIMPLE_TIME_FORMAT)


def string2bool(string_input):
    """ Convert given string into boolean value."""
    string_input = str(string_input).lower()
    return string_input in ("yes", "true", "t", "1")




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
"""

import os
import random
import shutil
import string
import uuid
import pyAesCrypt

from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile

LOGGER = get_logger(__name__)


class EncryptionHandler(object):
    encrypted_dir_name_regex = 'cipher_{}'
    encrypted_suffix = '.aes'

    def __init__(self, dir_gid):
        """
        :param dir_gid: the GID to use for the encrypted directory name
        """
        if isinstance(dir_gid, uuid.UUID):
            dir_gid = dir_gid.hex
        self.current_enc_dirname = self._prepare_encrypted_dir_name(dir_gid)
        self.enc_data_dir = TvbProfile.current.hpc.CRYPT_DATADIR
        self.pass_dir = TvbProfile.current.hpc.CRYPT_PASSDIR
        self.buffer_size = TvbProfile.current.hpc.CRYPT_BUFFER_SIZE
        self._generate_dirs()

    def _prepare_encrypted_dir_name(self, dir_gid):
        return self.encrypted_dir_name_regex.format(dir_gid)

    def _generate_dirs(self):
        for dirr in [self.enc_data_dir, self.pass_dir]:
            if not os.path.isdir(dirr):
                os.makedirs(dirr)

    def cleanup_encryption_handler(self):
        for path in [self.get_encrypted_dir(), self.get_password_file()]:
            if os.path.exists(path):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)

    @staticmethod
    def generate_random_password():
        pass_size = TvbProfile.current.hpc.CRYPT_PASS_SIZE
        chars = string.ascii_letters + string.digits
        password = ''.join(random.choice(chars) for i in range(pass_size))
        return password

    def _generate_password(self):
        password_file = self.get_password_file()
        if os.path.exists(password_file):
            return password_file
        password = self.generate_random_password()
        with open(password_file, 'w') as fd:
            fd.write(password)
        os.chmod(password_file, TvbProfile.current.ACCESS_MODE_TVB_FILES)
        return password_file

    @staticmethod
    def _read_password(password_file):
        with open(password_file, 'rb') as password_file:
            password = password_file.read().decode()
        return password

    def get_encrypted_dir(self):
        return os.path.join(self.enc_data_dir, self.current_enc_dirname)

    def get_password_file(self):
        return os.path.join(self.pass_dir, self.current_enc_dirname)

    def _prepare_encryption_dir(self, subdir):
        encrypted_dir = self.get_encrypted_dir()

        if subdir:
            encrypted_dir = os.path.join(encrypted_dir, subdir)

        if not os.path.isdir(encrypted_dir):
            os.makedirs(encrypted_dir)
        return encrypted_dir

    def encrypt_inputs(self, files_to_encrypt, subdir=None):
        # type: (list, str) -> list
        """
        Receive a list with all files to encrypt.
        Prepare encryption directory and encrypt each file.
        Return a list with all files from the encrypted directory.
        """
        encryption_dir = self._prepare_encryption_dir(subdir)
        password_file = self._generate_password()

        password = self._read_password(password_file)

        for file_to_encrypt in files_to_encrypt:
            encrypted_file = os.path.join(encryption_dir, os.path.basename(file_to_encrypt) + self.encrypted_suffix)
            pyAesCrypt.encryptFile(file_to_encrypt, encrypted_file, password, self.buffer_size)

        encrypted_files = [os.path.join(encryption_dir, enc_file) for enc_file in os.listdir(encryption_dir)]
        return encrypted_files

    def _determine_plain_filename(self, dir, encrypted_file):
        # type: (str, str) -> str
        return os.path.join(dir, os.path.basename(encrypted_file).replace(self.encrypted_suffix, ''))

    def decrypt_results_to_dir(self, dir, from_subdir=None):
        # type: (str, str) -> list
        """
        Having an already encrypted directory, decrypt all files,
        then move plain files to the location specified by :param dir
        """
        password = self._read_password(self.get_password_file())

        if not os.path.isdir(dir):
            os.makedirs(dir)

        encrypted_dir = self.get_encrypted_dir()

        if from_subdir:
            encrypted_dir = os.path.join(encrypted_dir, from_subdir)

        plain_files = []
        for encrypted_file in os.listdir(encrypted_dir):
            plain_file = self._determine_plain_filename(dir, encrypted_file)
            encrypted_file_full = os.path.join(encrypted_dir, encrypted_file)
            try:
                pyAesCrypt.decryptFile(encrypted_file_full, plain_file, password, self.buffer_size)
                plain_files.append(plain_file)
            except ValueError:
                LOGGER.info('Could not decrypt file {}'.format(encrypted_file))

        return plain_files

    def decrypt_files_to_dir(self, files, dir):
        # type: (list, str) -> list
        """
        Given a list of encrypted files, decrypt them,
        then move plain files to the location specified by :param dir
        """
        password = self._read_password(self.get_password_file())

        if not os.path.isdir(dir):
            os.makedirs(dir)

        plain_files = []
        for encrypted_file in files:
            plain_file = self._determine_plain_filename(dir, encrypted_file)
            try:
                pyAesCrypt.decryptFile(encrypted_file, plain_file, password, self.buffer_size)
                plain_files.append(plain_file)
            except ValueError:
                LOGGER.info('Could not decrypt file {}'.format(encrypted_file))

        return plain_files



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Bogdan Valean <bogdan.valean@codemart.ro>
"""

import os
import shutil
import threading
from io import BytesIO
from os import stat
from queue import Queue
from threading import Lock

import pyAesCrypt
import requests
from tvb.basic.exceptions import TVBException
from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.storage.h5.decorators import synchronized
from tvb.storage.h5.encryption.encryption_handler import EncryptionHandler
from tvb.storage.h5.file.files_helper import FilesHelper
from tvb.storage.kube.kube_notifier import KubeNotifier

LOGGER = get_logger(__name__)

try:
    from syncrypto import Crypto, Syncrypto
except ModuleNotFoundError:
    LOGGER.info("Cannot import syncrypto library.")


class InvalidStorageEncryptionException(TVBException):
    """
    Exception thrown when encryption storage cannot be allowed.
    """

    def __init__(self, message):
        TVBException.__init__(self, message)


class DataEncryptionHandlerMeta(type):
    """
    Metaclass used to generate the singleton instance
    """

    _instances = {}

    def __call__(cls):
        if cls._instances.get(cls, None) is None:
            cls._instances[cls] = super(DataEncryptionHandlerMeta, cls).__call__()
        return DataEncryptionHandlerMeta._instances[cls]


class DataEncryptionHandler(metaclass=DataEncryptionHandlerMeta):
    ENCRYPTED_FOLDER_SUFFIX = "_encrypted"
    KEYS_FOLDER = ".storage-keys"
    CRYPTO_PASS = "CRYPTO_PASS"
    APP_ENCRYPTION_HANDLER = "APP_ENCRYPTION_HANDLER"

    file_helper = FilesHelper()

    # Queue used to push projects which need synchronisation
    sync_project_queue = Queue()

    # Dict used to count the project frequency from the queue
    queue_elements_count = {}

    # Dict used to count how the project usage
    users_project_usage = {}

    running_operations = {}

    # Dict used to keep projects which are marked for deletion
    marked_for_delete = set()

    # Linked projects
    linked_projects = {}

    lock = Lock()

    def _queue_count(self, folder):
        return self.queue_elements_count[folder] if folder in self.queue_elements_count else 0

    def _project_active_count(self, folder):
        return self.users_project_usage[folder] if folder in self.users_project_usage else 0

    def _running_op_count(self, folder):
        return self.running_operations[folder] if folder in self.running_operations else 0

    @synchronized(lock)
    def inc_project_usage_count(self, folder):
        count = self._project_active_count(folder)
        count += 1
        self.users_project_usage[folder] = count

    @synchronized(lock)
    def _dec_project_usage_count(self, folder):
        count = self._project_active_count(folder)
        if count == 1:
            self.users_project_usage.pop(folder)
            return
        count -= 1
        self.users_project_usage[folder] = count

    @synchronized(lock)
    def inc_running_op_count(self, folder):
        count = self._running_op_count(folder)
        count += 1
        self.running_operations[folder] = count

    @synchronized(lock)
    def dec_running_op_count(self, folder):
        count = self._running_op_count(folder)
        if count == 1:
            self.running_operations.pop(folder)
            return
        count -= 1
        self.running_operations[folder] = count

    @synchronized(lock)
    def _inc_queue_count(self, folder):
        count = self._queue_count(folder)
        count += 1
        self.queue_elements_count[folder] = count

    @synchronized(lock)
    def dec_queue_count(self, folder):
        count = self._queue_count(folder)
        if count == 1:
            self.queue_elements_count.pop(folder)
            return
        count -= 1
        self.queue_elements_count[folder] = count

    @synchronized(lock)
    def check_and_delete(self, folder):
        # Check if we can remove a folder:
        #   1. It is not in the queue
        #   2. It is marked for delete
        #   3. Nobody is using it
        if not self.is_in_usage(folder) \
                and folder in self.marked_for_delete:
            self.marked_for_delete.remove(folder)
            LOGGER.info("Remove folder {}".format(folder))
            shutil.rmtree(folder)

    def is_in_usage(self, folder):
        return self._queue_count(folder) > 0 \
               or self._project_active_count(folder) > 0 \
               or self._running_op_count(folder) > 0

    @staticmethod
    def compute_encrypted_folder_path(current_project_folder):
        project_name = os.path.basename(current_project_folder)
        project_path = os.path.join(TvbProfile.current.TVB_STORAGE, FilesHelper.PROJECTS_FOLDER, project_name)
        return "{}{}".format(project_path, DataEncryptionHandler.ENCRYPTED_FOLDER_SUFFIX)

    def sync_folders(self, folder):
        if not self.encryption_enabled():
            return

        project_name = os.path.basename(folder)
        encrypted_folder = self.compute_encrypted_folder_path(folder)

        if os.path.exists(encrypted_folder) or os.path.exists(folder):
            crypto_pass = self._project_key(project_name)
            crypto = Crypto(crypto_pass)
            syncro = Syncrypto(crypto, encrypted_folder, folder)
            syncro.sync_folder()
            trash_path = os.path.join(encrypted_folder, "_syncrypto", "trash")
            if os.path.exists(trash_path):
                shutil.rmtree(trash_path)
        else:
            LOGGER.info("Project {} was deleted".format(project_name))

    def set_project_active(self, project, linked_dt):
        if not self.encryption_enabled():
            return
        project_folder = self.file_helper.get_project_folder(project.name)
        projects = set()
        if linked_dt is None:
            linked_dt = []
        for dt_path in linked_dt:
            project_path = FilesHelper.get_project_folder_from_h5(dt_path)
            projects.add(project_path)
        if len(linked_dt) > 0:
            self.linked_projects[project_folder] = projects
        projects.add(project_folder)

        for project_folder in projects:
            self.inc_project_usage_count(project_folder)
            self.push_folder_to_sync(project_folder)

    def set_project_inactive(self, project):
        if not self.encryption_enabled():
            return
        project_folder = self.file_helper.get_project_folder(project)
        projects = self.linked_projects.pop(project_folder) if project_folder in self.linked_projects else set()
        projects.add(project_folder)
        for project_folder in projects:
            self._dec_project_usage_count(project_folder)
            if self._queue_count(project_folder) > 0 \
                    or self._project_active_count(project_folder) > 0 \
                    or self._running_op_count(project_folder) > 0:
                self.marked_for_delete.add(project_folder)
                LOGGER.info("Project {} still in use. Marked for deletion.".format(project_folder))
                continue
            LOGGER.info("Remove project: {}".format(project_folder))
            shutil.rmtree(project_folder)

    def push_folder_to_sync(self, folder):
        if not self.encryption_enabled() or self._queue_count(folder) > 0:
            return
        self._inc_queue_count(folder)
        self.sync_project_queue.put(folder)

    @staticmethod
    def encryption_enabled():
        if not TvbProfile.current.web.ENCRYPT_STORAGE:
            return False
        if not TvbProfile.current.web.CAN_ENCRYPT_STORAGE:
            raise InvalidStorageEncryptionException(
                "We can not enable STORAGE ENCRYPTION. Most probably syncrypto is not installed!")
        return True

    @staticmethod
    def app_encryption_handler():
        app_encryption_handler = True if DataEncryptionHandler.APP_ENCRYPTION_HANDLER in os.environ and os.environ[
            DataEncryptionHandler.APP_ENCRYPTION_HANDLER].lower() == 'true' else False
        return not TvbProfile.current.web.OPENSHIFT_DEPLOY or app_encryption_handler

    @staticmethod
    def _get_unencrypted_projects(projects_folder):
        project_list = os.listdir(projects_folder)
        return list(map(lambda project: os.path.join(projects_folder, str(project)),
                        filter(lambda project: not str(project).endswith(DataEncryptionHandler.ENCRYPTED_FOLDER_SUFFIX),
                               project_list)))

    def startup_cleanup(self):
        unencrypted_projects_folder = os.path.join(TvbProfile.current.TVB_STORAGE, FilesHelper.PROJECTS_FOLDER)
        projects_list = self._get_unencrypted_projects(unencrypted_projects_folder)
        for project in projects_list:
            LOGGER.info("Sync and clean project: {}".format(project))
            self.sync_folders(project)
            shutil.rmtree(project)

    @staticmethod
    def project_key_path(project_name):
        return os.path.join(TvbProfile.current.TVB_STORAGE, DataEncryptionHandler.KEYS_FOLDER, str(project_name))

    @staticmethod
    def _project_key(project_name):
        password_encryption_key = os.environ[
            DataEncryptionHandler.CRYPTO_PASS] if DataEncryptionHandler.CRYPTO_PASS in os.environ else None
        if password_encryption_key is None:
            raise TVBException("Password encryption key is not defined.")
        project_keys_folder = os.path.join(TvbProfile.current.TVB_STORAGE, DataEncryptionHandler.KEYS_FOLDER)
        DataEncryptionHandler.file_helper.check_created(project_keys_folder)

        encrypted_project_key = DataEncryptionHandler.project_key_path(project_name)
        if os.path.exists(encrypted_project_key):
            with open(encrypted_project_key, "rb") as fIn:
                inputFileSize = stat(encrypted_project_key).st_size
                pass_stream = BytesIO()
                pyAesCrypt.decryptStream(fIn, pass_stream, password_encryption_key, 64 * 1024, inputFileSize)
                project_key = pass_stream.getvalue().decode()
                pass_stream.close()

            return project_key

        project_key = EncryptionHandler.generate_random_password()
        with open(encrypted_project_key, "wb") as fOut:
            pass_stream = BytesIO(str.encode(project_key))
            pyAesCrypt.encryptStream(pass_stream, fOut, password_encryption_key, 64 * 1024)
            pass_stream.close()
        return project_key


class FoldersQueueConsumer(threading.Thread):
    was_processing = False

    marked_stop = False

    def mark_stop(self):
        self.marked_stop = True

    def run(self):
        if not encryption_handler.encryption_enabled():
            return
        while True:
            if encryption_handler.sync_project_queue.empty():
                if self.was_processing:
                    self.was_processing = False
                    LOGGER.info("Finish processing queue")
                if self.marked_stop:
                    break
                continue
            if not self.was_processing:
                LOGGER.info("Start processing queue")
                self.was_processing = True
            folder = encryption_handler.sync_project_queue.get()
            encryption_handler.sync_folders(folder)
            encryption_handler.dec_queue_count(folder)
            encryption_handler.check_and_delete(folder)
            encryption_handler.sync_project_queue.task_done()


class DataEncryptionRemoteHandler(DataEncryptionHandler):
    lock = Lock()

    @staticmethod
    def _notify_pods(method, folder=None, **kwargs):
        if folder:
            kwargs['folder'] = folder
        if TvbProfile.current.web.OPENSHIFT_DATA_ENCRYPTION_HANDLER_APPLICATION == "":
            raise TVBException("Openshift Data Encryption handler application is not defined")
        openshift_pods = KubeNotifier.get_pods(TvbProfile.current.web.OPENSHIFT_DATA_ENCRYPTION_HANDLER_APPLICATION)
        if len(openshift_pods) == 0:
            raise TVBException("Openshift Data Encryption handler app not found")
        encryption_app = openshift_pods[0]
        auth_header = KubeNotifier.get_authorization_header()
        url = "http://{}:{}/kube/data_encryption_handler/{}".format(encryption_app.ip, str(TvbProfile.current.web.SERVER_PORT), method)
        return requests.post(url=url, headers=auth_header, data=kwargs)

    @synchronized(lock)
    def inc_project_usage_count(self, folder):
        self._notify_pods("inc_project_usage_count", folder)

    @synchronized(lock)
    def check_and_delete(self, folder):
        self._notify_pods("check_and_delete", folder)

    @synchronized(lock)
    def dec_queue_count(self, folder):
        self._notify_pods("dec_queue_count", folder)

    @synchronized(lock)
    def dec_running_op_count(self, folder):
        self._notify_pods("dec_running_op_count", folder)

    @synchronized(lock)
    def inc_project_usage_count(self, folder):
        self._notify_pods("inc_project_usage_count", folder)

    @synchronized(lock)
    def inc_running_op_count(self, folder):
        self._notify_pods("inc_running_op_count", folder)

    @synchronized(lock)
    def is_in_usage(self, folder):
        response = self._notify_pods("is_in_usage", folder)
        in_usage = True
        if response.ok and response.content.decode('utf-8').lower() == 'false':
            in_usage = False
        return in_usage

    def push_folder_to_sync(self, folder):
        self._notify_pods("push_folder_to_sync", folder)

    def startup_cleanup(self):
        self._notify_pods("startup_cleanup")

    def sync_folders(self, folder):
        self._notify_pods("sync_folders", folder)

    def set_project_active(self, project, linked_dt):
        self._notify_pods("set_project_active", **{'project': project, 'linked_dt': linked_dt})

    def set_project_inactive(self, project):
        self._notify_pods("set_project_inactive", **{'project': project})


class DataEncryptionHandlerBuilder:
    @staticmethod
    def build_handler():
        if DataEncryptionHandler.app_encryption_handler():
            return DataEncryptionHandler()
        return DataEncryptionRemoteHandler()


encryption_handler = DataEncryptionHandlerBuilder.build_handler()



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Service layer used for kubernetes calls.

.. moduleauthor:: Bogdan Valean <bogdan.valean@codemart.ro>
"""

import requests
from kubernetes import config, client
from kubernetes.config import incluster_config
from concurrent.futures.thread import ThreadPoolExecutor
from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile

LOGGER = get_logger(__name__)


class KubeNotifier(object):

    @staticmethod
    def get_pods(application):
        openshift_pods = None

        try:
            response = KubeNotifier.fetch_endpoints(application)
            openshift_pods = response[0].subsets[0].addresses

        except Exception as e:
            LOGGER.error("Failed to retrieve openshift pods for application {}".format(application), e)

        return openshift_pods

    @staticmethod
    def notify_pods(url, target_application=TvbProfile.current.web.OPENSHIFT_APPLICATION):

        if not TvbProfile.current.web.OPENSHIFT_DEPLOY:
            return

        LOGGER.info("Notify all pods with url {}".format(url))
        openshift_pods = KubeNotifier.get_pods(target_application)
        url_pattern = "http://{}:" + str(TvbProfile.current.web.SERVER_PORT) + url
        auth_header = KubeNotifier.get_authorization_header()

        with ThreadPoolExecutor(max_workers=len(openshift_pods)) as executor:
            for pod in openshift_pods:
                pod_ip = pod.ip
                LOGGER.info("Notify pod: {}".format(pod_ip))
                executor.submit(requests.get, url=url_pattern.format(pod_ip), headers=auth_header)

    @staticmethod
    def fetch_endpoints(target_application=TvbProfile.current.web.OPENSHIFT_APPLICATION):
        config.load_incluster_config()

        v1 = client.CoreV1Api()
        response = v1.read_namespaced_endpoints_with_http_info(target_application,
                                                               TvbProfile.current.web.OPENSHIFT_NAMESPACE)
        LOGGER.info(f"This is the response from KubeClient: {response}")
        return response

    @staticmethod
    def get_authorization_token():
        kube_config = incluster_config.InClusterConfigLoader(
            token_filename=incluster_config.SERVICE_TOKEN_FILENAME,
            cert_filename=incluster_config.SERVICE_CERT_FILENAME,
            try_refresh_token=True)
        kube_config.load_and_set(None)
        return kube_config.token

    @staticmethod
    def get_authorization_header():
        token = KubeNotifier.get_authorization_token()
        return {"Authorization": "{}".format(token)}

    @staticmethod
    def check_token(authorization_token):
        expected_token = KubeNotifier.get_authorization_token()
        assert authorization_token == expected_token



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import os
import json

from tvb.storage.h5.file.xml_metadata_handlers import XMLReader, XMLWriter


class TestMetaDataReadXML:
    """
    Tests for tvb.core.entities.file.metadatahandler.XMLReader class.
    """   
    TO_BE_READ_FILE = "test_read.xml"
    #Values expected to be read from file
    EXPECTED_DICTIONARY = {'status': "5-FINISHED",
                           'gid': '497b3d59-b3c1-11e1-b2e4-68a86d1bd4fa',
                           'user_group': 'cff_74',
                           'fk_from_algo': json.dumps({'classname': 'CFF_Importer', 'identifier': None,
                                                       'module': 'tvb.adapters.uploaders.cff_importer'})
                           }
        
    def setup_method(self):
        """
        Sets up necessary files for the tests.
        """
        self.file_path = os.path.join(os.path.dirname(__file__), 
                                      self.TO_BE_READ_FILE)
        self.meta_reader = XMLReader(self.file_path)
    
    def test_read_metadata(self):
        """
        Test that content return by read_metadata matches the
        actual content of the XML.
        """
        meta_data = self.meta_reader.read_metadata_from_xml()
        for key, value in self.EXPECTED_DICTIONARY.items():
            found_value = meta_data[key]
            assert value == found_value
        

class TestMetaDataWriteXML:
    """
    Tests for XMLWriter.
    """ 
    WRITABLE_METADATA = TestMetaDataReadXML.EXPECTED_DICTIONARY
    
    def setup_method(self):
        """
        Sets up necessary files for the tests.
        """
        self.meta_writer = XMLWriter(self.WRITABLE_METADATA)
        self.result_path = os.path.join(os.path.dirname(__file__), "Operation.xml")
      
    def teardown_method(self):
        """
        Remove created XML file.
        """
        os.remove(self.result_path)
        
    def test_write_metadata(self):
        """
        Test that an XML file is created and correct data is written in it.
        """
        assert not os.path.exists(self.result_path)
        self.meta_writer.write_metadata_in_xml(self.result_path)
        assert os.path.exists(self.result_path)
        reader = XMLReader(self.result_path)
        meta_data = reader.read_metadata_from_xml()
        for key, value in TestMetaDataReadXML.EXPECTED_DICTIONARY.items():
            found_value = meta_data[key]
            assert value == found_value



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Robert Vincze <robert.vincze@codemart.ro>
"""

import os
import pytest

from tvb.basic.profile import TvbProfile
from tvb.storage.h5.file.exceptions import FileStructureException
from tvb.storage.h5.file.files_helper import FilesHelper
from tvb.storage.h5.file.xml_metadata_handlers import XMLReader
from tvb.storage.storage_interface import StorageInterface
from tvb.tests.storage.dummy.dummy_project import DummyProject
from tvb.tests.storage.dummy.dummy_storage_data_h5 import DummyStorageDataH5
from tvb.tests.storage.storage_test import StorageTestCase

root_storage = TvbProfile.current.TVB_STORAGE


class TestFilesHelper(StorageTestCase):
    """
    This class contains tests for the tvb.storage.h5.file.files_helper module.
    """

    def storage_setup_method(self):
        self.files_helper = FilesHelper()
        self.project_name = "test_proj"

    def storage_teardown_method(self):
        self.delete_projects_folders()

    def test_check_created(self):
        """ Test standard flows for check created. """
        self.files_helper.check_created(TvbProfile.current.TVB_STORAGE)
        assert os.path.exists(root_storage), "Storage not created!"

        self.files_helper.check_created(os.path.join(root_storage, "test"))
        assert os.path.exists(root_storage), "Storage not created!"
        assert os.path.exists(os.path.join(root_storage, "test")), "Test directory not created!"

    def test_get_project_folder(self):
        """
        Test the get_project_folder method which should create a folder in case
        it doesn't already exist.
        """
        project_path = self.files_helper.get_project_folder(self.project_name)
        assert os.path.exists(project_path), "Folder doesn't exist"

        folder_path = self.files_helper.get_project_folder(self.project_name, "43")
        assert os.path.exists(project_path), "Folder doesn't exist"
        assert os.path.exists(folder_path), "Folder doesn't exist"

    def test_rename_project_structure(self):
        """ Try to rename the folder structure of a project. Standard flow. """
        self.files_helper.get_project_folder(self.project_name)
        path, name = self.files_helper.rename_project_structure(self.project_name, "new_name")
        assert path != name, "Rename didn't take effect."

    def test_rename_structure_same_name(self):
        """ Try to rename the folder structure of a project. Same name. """
        self.files_helper.get_project_folder(self.project_name)

        with pytest.raises(FileStructureException):
            self.files_helper.rename_project_structure(self.project_name, self.project_name)

    def test_write_project_metadata(self):
        """  Write XML for test-project. """
        user_id = 1
        dummy_project = DummyProject(self.project_name, "description", 3, user_id)
        self.files_helper.write_project_metadata(dummy_project.to_dict(), StorageInterface.TVB_PROJECT_FILE)
        expected_file = self.files_helper.get_project_meta_file_path(self.project_name,
                                                                     StorageInterface.TVB_PROJECT_FILE)
        assert os.path.exists(expected_file)
        project_meta = XMLReader(expected_file).read_metadata_from_xml()
        loaded_project = DummyProject(None, None, None, None)
        loaded_project.from_dict(project_meta, user_id)
        assert dummy_project.name == loaded_project.name
        assert dummy_project.description == loaded_project.description
        assert dummy_project.gid == loaded_project.gid
        expected_dict = dummy_project.to_dict()
        del expected_dict['last_updated']
        found_dict = loaded_project.to_dict()
        del found_dict['last_updated']
        self._dictContainsSubset(expected_dict, found_dict)
        self._dictContainsSubset(found_dict, expected_dict)

    def test_move_datatype(self):
        """
        Make sure associated H5 file is moved to a correct new location.
        """
        path = self.files_helper.get_project_folder(self.project_name)
        old_h5_path = os.path.join(path, "dummy_datatype.h5")
        DummyStorageDataH5(old_h5_path)
        assert os.path.exists(old_h5_path), "Test file was not created!"
        self.files_helper.move_datatype(self.project_name + '2', "1", old_h5_path)

        assert not os.path.exists(old_h5_path), "Test file was not moved!"
        new_file_path = os.path.join(self.files_helper.get_project_folder(self.project_name + '2', "1"),
                                     os.path.basename(old_h5_path))
        assert os.path.exists(new_file_path), "Test file was not created!"

    def test_remove_files_valid(self):
        """
        Pass a valid list of files and check they are all removed.
        """
        file_list = ["test1", "test2", "test3"]
        for file_n in file_list:
            fp = open(file_n, 'w')
            fp.write('test')
            fp.close()
        for file_n in file_list:
            assert os.path.isfile(file_n)
        self.files_helper.remove_files(file_list, False)
        for file_n in file_list:
            assert not os.path.isfile(file_n)

    def test_remove_folder(self):
        """
        Pass an open file pointer, but ignore exceptions.
        """
        folder_name = "test_folder"
        os.mkdir(folder_name)
        assert os.path.isdir(folder_name), "Folder should be created."
        self.files_helper.remove_folder(folder_name, False)
        assert not os.path.isdir(folder_name), "Folder should be deleted."

    def test_remove_folder_non_existing_ignore_exc(self):
        """
        Pass an open file pointer, but ignore exceptions.
        """
        folder_name = "test_folder"
        assert not os.path.isdir(folder_name), "Folder should not exist before call."
        self.files_helper.remove_folder(folder_name, ignore_errors=True)

    def test_remove_folder_non_existing(self):
        """
        Pass an open file pointer, but ignore exceptions.
        """
        folder_name = "test_folder"
        assert not os.path.isdir(folder_name), "Folder should not exist before call."
        with pytest.raises(FileStructureException):
            self.files_helper.remove_folder(folder_name, ignore_errors=False)

    def _dictContainsSubset(self, expected, actual, msg=None):
        """Checks whether actual is a superset of expected."""
        missing = []
        mismatched = []
        for key, value in expected.items():
            if key not in actual:
                return False
            elif value != actual[key]:
                return False
        return True



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Robert Vincze <robert.vincze@codemart.ro>
"""

import os
import shutil
from functools import wraps
from types import FunctionType
import decorator

from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.storage.storage_interface import StorageInterface

LOGGER = get_logger(__name__)


class BaseStorageTestCase(object):

    @staticmethod
    def delete_projects_folders():
        projects_folder = os.path.join(TvbProfile.current.TVB_STORAGE, StorageInterface.PROJECTS_FOLDER)
        if os.path.exists(projects_folder):
            for current_file in os.listdir(projects_folder):
                full_path = os.path.join(TvbProfile.current.TVB_STORAGE, StorageInterface.PROJECTS_FOLDER, current_file)
                if os.path.isdir(full_path):
                    shutil.rmtree(full_path, ignore_errors=True)


def storage_test(func, callback=None):
    """
    A decorator to be used in tests which makes sure all stored files and folders are deleted at the end of the test.
    """
    if func.__name__.startswith('test_'):

        @wraps(func)
        def dec(func, *args, **kwargs):
            try:
                if hasattr(args[0], 'storage_setup_method_TVB'):
                    LOGGER.debug(args[0].__class__.__name__ + "->" + func.__name__
                                 + "- Storage SETUP starting...")
                    args[0].storage_setup_method_TVB()
                result = func(*args, **kwargs)
            finally:
                if hasattr(args[0], 'storage_teardown_method_TVB'):
                    LOGGER.debug(args[0].__class__.__name__ + "->" + func.__name__
                                 + "- Storage TEARDOWN starting...")
                    args[0].storage_teardown_method_TVB()

            if callback is not None:
                callback(*args, **kwargs)
            return result

        return decorator.decorator(dec, func)
    else:
        return func


class StorageTestMeta(type):

    def __new__(mcs, classname, bases, class_dict):
        """
        Called when a new class gets instantiated.
        """
        new_class_dict = {}
        for attr_name, attribute in class_dict.items():
            if (type(attribute) == FunctionType and not (attribute.__name__.startswith('__')
                                                         and attribute.__name__.endswith('__'))):
                if attr_name.startswith('test_'):
                    attribute = storage_test(attribute)
                if attr_name in ('storage_setup_method', 'storage_teardown_method'):
                    new_class_dict[attr_name + '_TVB'] = attribute
                else:
                    new_class_dict[attr_name] = attribute
            else:
                new_class_dict[attr_name] = attribute
        return type.__new__(mcs, classname, bases, new_class_dict)


class StorageTestCase(BaseStorageTestCase, metaclass=StorageTestMeta):
    """
    This class makes sure that any test case that has 'storage_setup_method' and
    'storage_teardown_method' will have those methods called before each, and
    after each method respectively.
    """